import numpy as np
import pandas as pd
from scipy.stats import logistic
from pkg_resources import resource_filename


def load_sample_data(timevary):
    """Load data that is part of the zepid package. This data set comes from simulated data from Jessie Edwards
    (thanks Jess!). This data is used for examples on zepid.readthedocs

    Parameters
    --------------
    timevary : bool
        Whether to return the time-varying data set or the time fixed. If True then returns data set with repeated
        visits. If False then a data set with single observation per subject representing the 45-week risk is returned

    Notes
    -------------

    For the time-varying data set, the following variables are returned;
        * id - participant unique ID
        * enter - start of follow-up period
        * out - end of time period
        * male - indicator variable for male (1 = yes)
        * age0 - age at enter = 0
        * cd40 - CD4 T cell count at enter = 0
        * dvl0 - detectable viral load data at enter = 0
        * cd4 - CD4 T cell count at enter = t
        * dvl - viral load at enter = t
        * art - indicator of whether ART was prescribed at enter = t
        * drop - indicator of whether individual dropped out of the study at enter = t (1 = yes)
        * dead - indicator for death at out = t (1 = yes)

    For the time-fixed data set, the following variables are returned
        * id - participant unique ID
        * male - indicator variable for male (1 = yes)
        * age0 - age at enter = 0
        * cd40 - CD4 T cell count at enter = 0
        * dvl0 - detectable viral load data at enter = 0
        * art - indicator of whether ART was prescribed at enter = 0
        * t - total time contributed

    Returns
    ----------
    DataFrame
        Returns either a time-varying or time-fixed pandas DataFrame

    Examples
    --------
    Load the time-fixed exposure data set

    >>> from zepid import load_sample_data
    >>> load_sample_data(timevary=False)

    Load the time-varying exposure data set

    >>> load_sample_data(timevary=True)
    """
    cols = ['id', 'enter', 'out', 'male', 'age0', 'cd40', 'dvl0', 'cd4', 'dvl', 'art', 'drop', 'dead']
    df = pd.read_csv(resource_filename('zepid', 'datasets/data.dat'),
                     delim_whitespace=True, header=None, names=cols, index_col=False)
    df.sort_values(by=['id', 'enter'], inplace=True)
    if timevary:
        return df
    else:
        dfi = df.loc[df.groupby('id').cumcount() == 0, ['id', 'male', 'age0', 'cd40', 'dvl0', 'art']].copy()
        dfo = df.loc[df.id != df.id.shift(-1)][['id', 'dead', 'drop', 'out']].copy()
        dfo.loc[(dfo['drop'] == 1) & (dfo['out'] <= 45), 'dead'] = np.nan
        dfo['dead'] = np.where((dfo['dead'] == 1) & (dfo['out'] > 45), 0, dfo['dead'])
        dff = pd.merge(dfi, dfo, left_on='id', right_on='id')
        dff.rename(columns={'out': 't'}, inplace=True)
        dff.drop('drop', axis=1, inplace=True)

        # Adding last CD4 T-cell count for continuous outcome data
        dfl = df.loc[df['out'] <= 45, ['id', 'cd4']].copy()
        dfll = dfl.loc[dfl.id != dfl.id.shift(-1)].copy()
        dfll.rename(columns={'cd4': 'cd4_wk45'}, inplace=True)
        dff = pd.merge(dff, dfll, left_on='id', right_on='id')
        dff['cd4_wk45'] = np.where(dff['dead'] == 1, np.nan, dff['cd4_wk45'])
        dff['t'] = np.where(dff['t'] >= 45, 45, dff['t'])

        return dff


def load_ewing_sarcoma_data():
    """Loads Ewing's Sarcoma survival data from Glaubiger DL, Makuch R, Schwarz J, Levine AS, Johnson RE. Determination
    of prognostic factors and their influence on therapeutic results in patients with Ewing's sarcoma. Cancer.
    1980;45(8):2213-9. In total, data for 76 patients is loaded. Further descriptions of the 76 patients and the data
    can be found in Makuch RW. Adjusted survival curve estimation using covariates. J Chronic Dis. 1982;35(6):437-43.
    or Cole SR, Hernán MA. Adjusted survival curves with inverse probability weights. Comput Methods Programs
    Biomed. 2004;75(1):45-9.

    Notes
    -----------
    Variables within the dataset are
        * treat - treatment (1 is novel treatment; 0 is one of three standard treatments)
        * ldh - pre-treatment serum lactic acid dehydrogenase (LDH) (1 is >= 200 international units; 0 is <200)
        * time - days till recurrence or censoring (continuous)
        * outcome - sarcoma recurrence (1 is recurrence; 0 is censored)

    Returns
    ----------
    DataFrame
        Returns pandas DataFrame
    """
    return pd.read_csv(resource_filename('zepid', 'datasets/ewing.dat'), index_col=False)


def load_gvhd_data():
    """Loads bone marrow transplant recipient data from Keil AP, Edwards JK, Richardson DB, Naimi AI, Cole SR. The
    parametric g-formula for time-to-event data: intuition and a worked example. Epidemiology. 2014;25(6):889-97.
    Patients were followed until death or administrative censoring at 5-years.

    Notes
    ---------------
    Variables are formatted exactly as described in Keil et al. 2014
        * id: unique ID for each participant
        * age: participant baseline age
        * agesq: squared baseline age
        * agecurs1: restricted cubic spline knot 1 for baseline age
        * agecurs2: restricted cubic spline knot 2 for basline age
        * male: participant gender (1 is male, 0 is female)
        * cmv: cytomegalovirus baseline immune status (1 is yes, 0 is no)
        * all: at this time, I am unsure what this variable indicates (1, 0)
        * wait: wait time from diagnosis to transplantation (months)
        * day: day since transplantation
        * daysq: squared day since transplantation
        * daycu: cubic day since transplantation
        * daycurs1: restricted cubic spline knot 1 for days since transplantation
        * daycurs2: restricted cubic spline knot 2 for days since transplantation
        * yesterday: previous day
        * tomorrow: day after
        * gvhd: indicator for Graph-versus-Host Disease (1 is yes, 0 is no)
        * d: indicator of death (1 is yes, 0 is no)
        * relapse: indicator for relapse (1 is yes, 0 is no)
        * platnorm: indicator for normal platelet count (1 is yes, 0 is no)
        * censlost: indicator for censoring due to loss-to-follow-up (1 is yes, 0 is no)
        * gvhdm1: indicator for previous day diagnosis of GvHD (1 is yes, 0 is no)
        * relapsem1: indicator for previous day relapse (1 is yes, 0 is no)
        * platnormm1: indicator for previous day normal platelet count (1 is yes, 0 is no)
        * daysnogvhd: number of consecutive days without a GvHD diagnosis
        * daysnorelapse: number of consecutive days without relapse
        * daysnoplatnorm: number of consecutive days without normal platelet count
        * daysgvhd: number of consecutive days with GvHD
        * daysrelapse: number of consecutive days after relapse
        * daysplatnorm: number of consecutive days with normal platelet count

    Returns
    ----------
    DataFrame
        Returns pandas DataFrame
    """
    df = pd.read_csv(resource_filename('zepid', 'datasets/gvhd.dat'), delim_whitespace=True, index_col=False)

    # Coding variables to match Keil et al. exactly
    df['wait'] = df['waitdays'] / 30.5
    df['agesq'] = df['age']**2
    df['agecurs1'] = ((df['age'] > 17.0) * (df['age'] - 17.0) ** 3 -
                      ((df['age'] > 30.0) * (df['age'] - 30.0) ** 3) * (41.4 - 17.0) / (41.4 - 30.0))
    df['agecurs2'] = ((df['age'] > 25.4) * (df['age'] - 25.4) ** 3 -
                      ((df['age'] > 41.4) * (df['age'] - 41.4) ** 3) * (41.4 - 25.4) / (41.4 - 30.0))

    # Expanding data into person-time periods
    df['t_int'] = df['t'].astype(int)
    lf = pd.DataFrame(np.repeat(df.values, df['t_int'] + 1, axis=0), columns=df.columns)
    lf['yesterday'] = lf.groupby('id')['t'].cumcount()
    lf['day'] = lf['yesterday'] + 1
    lf['tomorrow'] = lf['day'] + 1
    lf['daysq'] = lf['day']**2
    lf['daycu'] = lf['day']**3
    lf['daycurs1'] = (((lf['day'] > 63) * ((lf['day'] - 63) / 63) ** 3) +
                      ((lf['day'] > 716) * ((lf['day'] - 716) / 63) ** 3) * (350.0 - 63) -
                      ((lf['day'] > 350) * ((lf['day'] - 350) / 63) ** 3) * (716 - 63) / (716 - 350))
    lf['daycurs2'] = (((lf['day'] > 168) * ((lf['day'] - 168) / 63) ** 3) +
                      ((lf['day'] > 716) * ((lf['day'] - 716) / 63) ** 3) * (350 - 168) -
                      ((lf['day'] > 350) * ((lf['day'] - 350) / 63) ** 3) * (716 - 168) / (716 - 350))
    lf['tdiff'] = lf['t'] - lf['yesterday']
    lf = lf.loc[lf['tdiff'] != 0].copy()  # Removes redundant first observation

    # Recoding variables for the expanded data
    lf['d'] = (lf['day'] >= lf['t']) * lf['d_dea']
    lf['gvhd'] = (lf['day'] > lf['t_gvhd']) * 1
    lf['relapse'] = (lf['day'] > lf['t_rel']) * 1
    lf['platnorm'] = (lf['day'] > lf['t_pla']) * 1
    lf['gvhdm1'] = (lf['yesterday'] > lf['t_gvhd']) * 1
    lf['relapsem1'] = (lf['yesterday'] > lf['t_rel']) * 1
    lf['platnormm1'] = (lf['yesterday'] > lf['t_pla']) * 1

    # Censoring time point
    lf['censeof'] = np.where((lf['day'] == lf['t']) & (lf['d'] == 0) & (lf['day'] == 1825), 1, 0)
    lf['censlost'] = np.where((lf['day'] == lf['t']) & (lf['d'] == 0) & (lf['day'] != 1825), 1, 0)

    # Setting initial values for days from X
    lf['daysnorelapse'] = np.where(lf['relapse'] == 0, lf['day'], np.nan)
    lf['daysnorelapse'] = lf['daysnorelapse'].fillna(method='ffill')
    lf['daysnoplatnorm'] = np.where(lf['id'] != lf['id'].shift(1), 0, np.nan)
    lf['daysnoplatnorm'] = np.where(lf['platnorm'] == 0, lf['day'], lf['daysnoplatnorm'])
    lf['daysnoplatnorm'] = lf['daysnoplatnorm'].fillna(method='ffill')
    lf['daysnogvhd'] = np.where(lf['gvhd'] == 0, lf['day'], np.nan)
    lf['daysnogvhd'] = lf['daysnogvhd'].fillna(method='ffill')

    lf['daysrelapse'] = lf.groupby('id')['relapse'].cumsum()
    lf['daysplatnorm'] = lf.groupby('id')['platnorm'].cumsum()
    lf['daysgvhd'] = lf.groupby('id')['gvhd'].cumsum()
    return lf[['id', 'age', 'agesq', 'agecurs1', 'agecurs2', 'male', 'cmv', 'all', 'wait', 'yesterday', 'tomorrow',
               'day', 'daysq', 'daycu', 'daycurs1', 'daycurs2', 'd', 'gvhd', 'relapse', 'platnorm', 'gvhdm1',
               'relapsem1', 'platnormm1', 'censlost', 'daysnorelapse', 'daysnoplatnorm', 'daysnogvhd',
               'daysrelapse', 'daysplatnorm', 'daysgvhd']]


def load_sciatica_data():
    """Loads the Sciatica Trial data published in; Mertens, BJA, Jacobs, WCH, Brand, R, and Peul, WC. Assessment of
    patient-specific surgery effect based on weighted estimation and propensity scoring in the re-analysis of the
    Sciatica Trial. PLOS One 2014. Details of the original Sciatica Trial are available in; Peul WC, van Houwelingen HC,
    van den Hout WB, et al. Surgery versus Prolonged Conservative Treatment for Sciatica. NEJM 2007
    DOI: 10.1177/0962280214545529

    Notes
    -------------
    Variables included are
        * id: unique identifier for patient
        * tpoints: follow-up time period
        * time: follow-up time
        * age_b: age at follow-up time
        * age_t: age at randomization
        * vas1_t: VAS score
        * vas2_t: VAS score
        * roland_t: Roland score
        * likert_t: Likert score
        * vas1_b: VAS1 at baseline
        * vas2_b: VAS2 at baseline
        * roland_b: Roland score at baseline
        * likert_b: Likert score at baseline
        * male: participant gender (1 is male, 0 is female)
        * weight: participant weight in kilograms
        * height: participant height in meters
        * surgery: whether participant received the surgery (1 is surgery, 0 is not yet surgery)

    Returns
    ----------
    DataFrame
        Returns pandas DataFrame
    """
    df = pd.read_csv(resource_filename('zepid', 'datasets/sciatica.dat'), delim_whitespace=True, index_col=False,
                     header=None, names=['id', 'tpoints', 'time', 'age_b', 'age_t', 'vas1_t', 'vas2_t', 'roland_t',
                                         'likert_t', 'vas1_b', 'vas2_b', 'roland_b', 'likert_b', 'male', 'weight',
                                         'height', 'surgery'])
    return df


def load_leukemia_data():
    """Loads data from Freireich EJ et al., "The Effect of 6-Mercaptopurine on the Duration of Steriod-induced
    Remissions in Acute Leukemia: A Model for Evaluation of Other Potentially Useful Therapy" Blood 1963

    Notes
    ------------
    Variables included are
        t: time
        status: event indicator (0: censored, 1: relapsed)
        sex: male, female
        logwbc: log-transformed white blood cell count
        treat: treatment indicator

    Returns
    ----------
    DataFrame
        Returns pandas DataFrame
    """
    df = pd.read_csv(resource_filename('zepid', 'datasets/leukemia.dat'), delim_whitespace=True, index_col=False)
    return df


def load_binge_drinking_data():
    """Loads data from Ahren J et al., "Predicting the Population Health Impacts of Community Interventions: The Case
    of Alcohol Outlets and Binge Drinking" AJPH 2016. Below is some notes taken from the supplementary materials
    detailed by the paper authors;

    "The data provided for use with this sample code are simulated. The data are designed to be similar to the real
    data and associations examined in the main paper. There are 4000 observations representing individuals who are
    nested in 44 communities (variable name: neighborhood_id). The exposure of interest is neighborhood alcohol outlet
    density (alc_outlet_density), with values ranging from 39 to 168. The outcome of interest is a  binary indicator of
    binge drinking (binge_drink), and covariates include gender (male), age (age_categorical), marital status
    (married), education (education_categorical), and race/ethnicity (race_categorical). Alcohol outlet density and
    binge drinking were simulated as simple linear functions of the covariates. Thus, unlike the applied example, the
    relation of outlet density with binge drinking has a linear shape."

    Notes
    -----
    Variables included are
        * male: gender (0: female, 1: male)
        * age_categorical: age groups (not clearly defined as to what they refer to)
        * married: marital status
        * education_categorical: categories of education levels
        * race_categorical: categories of race
        * alc_outlet_density: density of alcohol outlets in the neighborhood (continuous)
        * binge_drink: whether individual binge drinks (1: yes, 0: no)
        * neighborhood_id: identifier for groups that individuals are nested in

    Returns
    -------
    DataFrame
        Returns pandas DataFrame
    """
    df = pd.read_csv(resource_filename('zepid', 'datasets/binge.dat'), index_col=False)
    return df


def load_longitudinal_data():
    """Loads simulated longitudinal data. This longitudinal data is used to demonstrate the sequential regression
    time-varying g-formula, and the longitudinal targeted maximum likelihood estimator. The format of the returned
    file is a long data set

    Notes
    -----
    Variables included are
       * A: treatment of interest
       * Y: outcome of interest
       * t: time-point
       * W: baseline variable
       * L: time-varying variable
       * id: unique identifier for each subject

    Returns
    -------
    DataFrame
        Returns pandas DataFrame
    """
    df = pd.read_csv(resource_filename('zepid', 'datasets/longitudinal.dat'), index_col=False)
    return df


def load_case_control_data():
    """Loads example case-control data, meant to reflect observed data. The goal of this case-control is to try to
    determine the causative agent / route

    Notes
    -----
    Variables included are:
        * id: id number
        * case: whether individual is a case or control (0: control, 1:case)
        * breakfast: whether individual ate breakfast (0: no, 1: yes)
        * lunch: lunch eaten by individual (0: no lunch, 1: salad, 2: chicken)
        * snack: whether individual ate snack (0: no, 1:yes)
        * dinner: whether individual ate dinner (0: no, 1:yes)
        * sleep: hours of sleep last night (0: 6-8 hours, 1: 8+ hours)
        * sunscreen: sunscreen usage (0: no, 1: yes)
        * hand_wash: hand hygiene habits (1: none, 2: poor, 3: average, 4: high, 5: optimal)

    Returns
    -------
    DataFrame
        Returns pandas DataFrame
    """
    df = pd.read_csv(resource_filename('zepid', 'datasets/case_control.dat'), delim_whitespace=True, index_col=False)
    return df


def load_monotone_missing_data():
    """Loads simulated data for monotone missing. This data is for demonstration with inverse probability of missing
    weights with data that is has a monotone missing pattern.

    Notes
    -----
    Variables included are:
        * id: id number
        * A: variable with no missing data
        * B: variable with missing data, conditional on A and L
        * C: variable with missing data, conditional on B and L
        * L: variable related to the missingness mechanism

    The above data is monotone missing with all individuals missing C are also missing B.

    True means in the observed data:
        * B: 0.4258
        * C: 0.5343

    Returns
    -------
    DataFrame
        Returns pandas DataFrame
    """
    df = pd.read_csv(resource_filename('zepid', 'datasets/monotone.dat'), index_col=False)
    return df[['id', 'A', 'B', 'C', 'L']]


def load_generalize_data(confounding):
    """Loads simulated data for demonstration of generalizability/transportability methods.

    Notes
    -----
    For the data, the true effect measures are
        Generalizability
            RD: 0.04651
            RR: 1.147
        Transportability
            RD: 0.02887
            RR: 1.085

    Parameters
    ----------
    confounding : bool
        Whether to load data with a confounder (True) or data from a hypothetical RCT (False). If the confounder is
        loaded, then the method used will need to account for confounding

    Returns
    -------
    DataFrame
        Returns pandas DataFrame
    """
    if confounding:
        df = pd.read_csv(resource_filename('zepid', 'datasets/generalize_conf.dat'), index_col=False)
    else:
        df = pd.read_csv(resource_filename('zepid', 'datasets/generalize_rct.dat'), index_col=False)
    df['id'] = df.index
    return df[['id', 'Y', 'A', 'S', 'L', 'W']]


def load_zivich_breskin_data():
    """Loads the singular simulation data set from Zivich PN & Breskin 2021.

    Notes
    -----------
    Variables within the dataset are
        * Y - outcome (atherosclerotic cardiovascular disease)
        * statin - treatment (1 is given statins; 0 is not)
        * age - Age
        * ldl_log - log-transformed LDL
        * diabetes - diabetes indicator
        * risk_score - calculated risk score between 0 and 1

    Returns
    ----------
    DataFrame
        Returns pandas DataFrame
    """
    return pd.read_csv(resource_filename('zepid', 'datasets/zivich_breskin_sim.csv'), index_col=False)



from .utils import (risk_ci, risk_ratio, risk_difference, number_needed_to_treat, odds_ratio,
                    incidence_rate_ci, incidence_rate_ratio, incidence_rate_difference,
                    attributable_community_risk, population_attributable_fraction,
                    probability_to_odds, odds_to_probability, logit, inverse_logit, probability_bounds,
                    counternull_pvalue, s_value,
                    sensitivity, specificity, ppv_converter, npv_converter, screening_cost_analyzer,
                    rubins_rules, semibayes)



import matplotlib.pyplot as plt
import networkx as nx
from networkx.algorithms.dag import descendants, ancestors
from itertools import combinations


class DirectedAcyclicGraph:
    def __init__(self, exposure, outcome):
        """Constructs a Directed Acyclic Graph (DAG) for determination of adjustment sets

        Parameters
        ----------
        exposure : str
            Exposure of interest in the causal diagram
        outcome : str
            Outcome of interest in the causal diagram

        # TODO add other implementations in the future... have as self.mediator, self.censor, self.missing

        Examples
        --------
        Setting up environment

        >>> from zepid.causal.causalgraph import DirectedAcyclicGraph

        Creating directed acyclic graph

        >>> dag = DirectedAcyclicGraph(exposure="X", outcome="Y")
        >>> dag.add_arrow(source="X", endpoint="Y")
        >>> dag.add_arrow(source="V", endpoint="Y")
        >>> dag.add_arrows(pairs=(("W", "X"), ("W", "Y")))

        Determining adjustment sets

        >>> dag.calculate_adjustment_sets()
        >>> dag.adjustment_sets
        >>> dag.minimal_adjustment_sets

        Plot diagram

        >>> dag.draw_dag()
        >>> plt.show()

        Assess arrow misdirections that result in the chosen adjustment set being invalid

        >>> dag.assess_misdirections(chosen_adjustment_set=set("W"))

        References
        ----------
        Shrier I, & Platt RW. (2008). Reducing bias through directed acyclic graphs.
        BMC medical research methodology, 8(1), 70.
        """
        self.exposure = exposure
        self.outcome = outcome

        dag = nx.DiGraph()
        dag.add_edge(self.exposure, self.outcome)
        self.dag = dag

        self.adjustment_sets = None
        self.minimal_adjustment_sets = None
        self.arrow_misdirections = None

    def add_arrow(self, source, endpoint):
        """Add a single arrow to the current causal DAG

        Parameters
        ----------
        source : str
            Node that arrow originates from
        endpoint : str
            Node that arrow points to
        """
        dag = self.dag.copy()
        dag.add_edge(source, endpoint)
        if not nx.is_directed_acyclic_graph(dag):
            raise DAGError("Cyclic graph detected. Invalid addition for arrow.")

        self.dag = dag

    def add_arrows(self, pairs):
        """Add a set of arrows to the current causal DAG

        Parameters
        ----------
        pairs : list, set, container
            Set of sets of node pairs to add arrows in the DAG, with the first node as the source (node that arrows
             originates from) and the second node is the endpoint (node that arrow points to)
        """
        dag = self.dag.copy()
        dag.add_edges_from(pairs)
        if not nx.is_directed_acyclic_graph(dag):
            raise DAGError("Cyclic graph detected. Invalid addition for arrow(s).")

        self.dag = dag

    def add_from_networkx(self, network):
        # Checking that it is a directed acyclic graph
        if not nx.is_directed_acyclic_graph(network):
            raise DAGError("Cyclic graph detected. Invalid networkx input.")

        # Checking that exposure and outcome are valid nodes
        nodes = list(network.nodes)
        if self.exposure not in nodes:
            raise DAGError(str(self.exposure)+" is not a node in the DAG")
        if self.outcome not in nodes:
            raise DAGError(str(self.outcome)+" is not a node in the DAG")

        self.dag = network.copy()

    def draw_dag(self, positions=None, invert=False, fig_size=(6, 5), node_size=1000):
        """Draws the current input causal DAG

        Parameters
        ----------
        positions :
            Option to provide node locations based on other functionalities
        invert : bool, optional
            To display the (often hidden) assumptions of causal DAG, the invert options displays all arrows assumed
            to not exist. The reversal is an easy way to display lower densities of in the original DAG. No direction
            is provided for these arrows
        fig_size : set
            controls the figure size returned
        node_size : int, float
            Size of nodes in the subsequent plot
        """
        if invert:
            dag = nx.complement(self.dag.to_undirected())
        else:
            dag = self.dag.copy()

        fig = plt.figure(figsize=fig_size)
        ax = plt.subplot(1, 1, 1)

        if positions is None:
            positions = nx.spectral_layout(self.dag)

        nx.draw_networkx(dag, positions, node_color="#d3d3d3", node_size=node_size,
                         edge_color='black', linewidths=1.0, width=1.5,
                         arrowsize=15, ax=ax, font_size=12)
        plt.axis('off')
        return ax

    def calculate_adjustment_sets(self):
        """Determines all sufficient adjustment sets for the causal diagram using the algorithm described in Shrier &
        Platt "Reducing bias through directed acyclic graphs" BMC Medical Research Methodology 2008.

        All possible adjustment sets are enumerated and then assessed. We can briefly consider this as a backtracking
        algorithm where we assess each possible combination that exists within the data

        # TODO in future should allow for adjustment sets to determine causal, censor, missing sets. default to all

        Adjustment sets are added as `DirectedAcyclicGraph.adjustment_sets` and
        `DirectedAcyclicGraph.minimal_adjustment_sets`
        """
        # Extracting list of all sets to check
        sets_to_check = self._define_all_adjustment_sets_(dag=self.dag)

        valid_sets = []
        for adj_set in sets_to_check:
            if self._check_valid_adjustment_set_(graph=self.dag, adjustment_set=adj_set):
                valid_sets.append(adj_set)

        self.adjustment_sets = valid_sets
        self.minimal_adjustment_sets = [x for x in valid_sets if len(x) == len(min(valid_sets, key=len))]

    def _define_all_adjustment_sets_(self, dag):
        """Background function to determine all possible adjustment set combinations to explore. Used to explore every
        possible combinations of adjustment sets to assess whether they are valid for d-separation.
        """
        # List of all nodes valid for adjustment
        all_nodes = list(dag.nodes)
        all_nodes.remove(self.exposure)
        all_nodes.remove(self.outcome)
        list_of_sets = []
        for i in range(0, len(all_nodes) + 1):
            list_of_sets.extend([x for x in combinations(all_nodes, i)])
        return list_of_sets

    def _check_valid_adjustment_set_(self, graph, adjustment_set):
        """Checks the adjustment set as valid using the following 6 steps
        Step 1) check no descendants of X are included in adjustment set
        Step 2) delete variables that meet certain definitions
        Step 3) delete all arrows that originate from exposure
        Step 4) connect all source nodes (to assess for collider stratification)
        Step 5) convert to undirected graph
        Step 6) check whether a path exists between exposure & outcome
        """
        dag = graph.copy()

        # List of all nodes valid for adjustment
        all_nodes = list(dag.nodes())
        all_nodes.remove(self.exposure)
        all_nodes.remove(self.outcome)

        # Step 1) Check no descendants of X
        desc_x = descendants(dag, self.exposure)
        if desc_x & set(adjustment_set):
            return False

        # Step 2) Delete all variables that: (a) non-ancestors of X, (b) non-ancestors of Y, (c) non-ancestors
        #         of adjustment set
        set_check = set(adjustment_set).union([self.exposure, self.outcome])
        set_remove = set(dag.nodes)
        for n in set_check:
            set_remove = set_remove & (dag.nodes - ancestors(dag, n))
        set_remove = set_remove - set([self.exposure, self.outcome]) - set(adjustment_set)
        dag.remove_nodes_from(set_remove)

        # Step 3) Delete all arrows with X as the source
        for endpoint in list(dag.successors(self.exposure)):
            dag.remove_edge(self.exposure, endpoint)

        # Step 4) Directly connect all source nodes pointing to same endpoint (for collider assessment)
        for n in dag:
            sources = list(dag.predecessors(n))
            if len(sources) > 1:
                for s1, s2 in combinations(sources, 2):
                    if not (dag.has_edge(s2, s1) or dag.has_edge(s1, s2)):
                        dag.add_edge(s1, s2)

        # Step 5) Remove arrow directionality
        uag = dag.to_undirected()

        # Step 6) Remove nodes from the adjustment set
        uag.remove_nodes_from(adjustment_set)

        # Checking whether a a path between X and Y exists now
        if nx.has_path(uag, self.exposure, self.outcome):
            return False
        else:
            return True

    def assess_misdirections(self, chosen_adjustment_set):
        """Arrow direction can potentially be misspecified. This function checks every possible arrow reversal and
        determines the relevant sufficient adjustment sets. Those new adjustment sets are compared to the chosen
        adjustment set for differences.

        Parameters
        ----------
        chosen_adjustment_set : set, list, container
            The sufficient adjustment set selected for the data analysis
        """
        all_edges = list(self.dag.edges())
        l = []
        for i in range(0, len(all_edges)+1):
            l.append([x for x in combinations(all_edges, i)])

        valid_switches = []
        valid_graphs = []
        for c in range(1, len(l)):
            for s in l[c]:
                # Copy graph
                g = self.dag.copy()
                g.remove_edges_from(s)  # Remove edge
                # Reversing all edges in that set
                for pair in s:
                    g.add_edge(pair[1], pair[0])  # Add reversed edge
                # Check if DAG
                if nx.is_directed_acyclic_graph(g):
                    valid_graphs.append(g)
                    valid_switches.append(s)

        alternative_adjustment_sets = {}
        for v, g in zip(valid_switches, valid_graphs):
            sets_to_check = self._define_all_adjustment_sets_(dag=g)
            valid_sets = []
            for adj_set in sets_to_check:
                if self._check_valid_adjustment_set_(graph=g, adjustment_set=adj_set):
                    valid_sets.append(adj_set)
            if chosen_adjustment_set not in set(valid_sets):
                alternative_adjustment_sets[v] = valid_sets

        # print(alternative_adjustment_sets)
        self.arrow_misdirections = alternative_adjustment_sets


class DAGError(Exception):
    """Exception raised for errors in Directed Acyclic Graphs not being directed or acyclic
    """

    def __init__(self, message):
        super().__init__(message)



import warnings
import patsy
import numpy as np
import pandas as pd
import scipy.optimize

from zepid.calc.utils import probability_bounds
from zepid.causal.utils import propensity_score, check_input_data


class GEstimationSNM:
    r"""G-estimation for structural nested mean models. G-estimation is distinct from the other g-methods (inverse
    probability weights and g-formula) in the parameter it estimates. Rather than estimating the average causal effect
    of treating everyone versus treating no one, g-estimation estimates the average causal effect within strata of L.
    It does this by specifying a structural nested model. The structural nested mean model looks like the following for
    additive effects

    .. math::

        E[Y^a |A=a, V] - E[Y^{a=0}|A=a, V] = \psi a + \psi a*V

    There are two items to note in the structural nested model; (1) there is no intercept or term for L, and (2) we
    need the potential outcomes to solve for psi. The first item means that we are estimating fewer parameters, making
    g-estimation less susceptible to model misspecification than the g-formula. The second means we cannot solve the
    above equation directly.

    Under the assumption of conditional exchangeability, we can solve for psi using another equation. Specifically, we
    can work to solve the following model

    .. math::

        logit(\Pr(A=1|Y^{a=0}, L)) = alpha + alpha Y^{a=0} + alpha Y{a=0} V + alpha L

    Under the assumption of conditional exchangeability, the alpha term for the potential outcome Y should be equal to
    zero! Therefore, we need to find the value of psi that results in that alpha term equaling zero. For the additive
    model, we can solve for psi in the first equation by

    .. math::

        H(\psi) = Y - (\psi A + \psi A L)

    meaning we solve for when alpha is approximately zero under

    .. math::

        logit(\Pr(A=1|Y^{a=0}, L)) = alpha + alpha H(\psi) + alpha H(\psi) V + alpha L

    To find the values for the psi's where the alpha for those terms is approximately zero, we have two options;
    (1) grid-search or (2) closed form. The closed form is ultimately faster since we are only required to do some
    basic matrix manipulation to solve. For the grid search, we need to search across the potential values that
    minimize the values of alphas. We use SciPy's Nelder-Mead optimization procedure for the heavy lifting.

    Parameters
    ----------
    df : DataFrame
        Pandas DataFrame object containing all variables of interest
    exposure : str
        Column name of the exposure variable. Currently only binary is supported
    outcome : str
        Column name of the outcome variable. Either continuous or binary outcomes are supported
    weights :
        Column name of weights. Weights allow for items like sampling weights, missing weights, and censoring weights
        to estimate effects

    Notes
    -----
    Similar to marginal structural models, g-estimation cannot inherently account for missing at random data. To
    account for missing outcome data, inverse probability of missing weights should be used

    The grid-search approach does allow for some unique sensitivity analyses that are not incorporated into the
    closed-form. Specifically, we can imagine that there is some unobserved confounding. With unobserved confounding,
    we know that the alpha value will not exactly equal zero. We can optimize for slightly different alphas to see
    how sensitive our results are to some assumptions regarding unobserved confounding. For further details on
    translating unobserved confounding to alpha values, see Scharfstein et al. 1999 in the references

    If you continuous variable takes on large values, you may see the closed-form and grid-search start to diverge in
    results. This is because of the tolerance value. If you have large outcome values, I recommend rescaling them to
    prevent any issues with the grid-search

    Examples
    --------
    Set up the environment and the data set

    >>> from zepid import load_sample_data, spline
    >>> from zepid.causal.snm import GEstimationSNM
    >>> df = load_sample_data(timevary=False).drop(columns=['dead'])
    >>> df[['cd4_rs1','cd4_rs2']] = spline(df,'cd40',n_knots=3,term=2,restricted=True)
    >>> df[['age_rs1','age_rs2']] = spline(df,'age0',n_knots=3,term=2,restricted=True)

    One-parameter structural nested mean model via closed-form solution

    >>> snm = GEstimationSNM(df, exposure='art', outcome='cd4_wk45')
    >>> snm.exposure_model('male + age0 + age_rs1 + age_rs2 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')
    >>> snm.structural_nested_model(model='art')
    >>> snm.fit()
    >>> snm.summary()

    One-parameter structural nested mean model via grid-search

    >>> snm = GEstimationSNM(df, exposure='art', outcome='cd4_wk45')
    >>> snm.exposure_model('male + age0 + age_rs1 + age_rs2 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')
    >>> snm.structural_nested_model(model='art')
    >>> snm.fit(solver='search')

    One-parameter structural nested mean model via grid-search with different alphas

    >>> snm = GEstimationSNM(df, exposure='art', outcome='cd4_wk45')
    >>> snm.exposure_model('male + age0 + age_rs1 + age_rs2 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')
    >>> snm.structural_nested_model(model='art')
    >>> snm.fit(solver='search', alpha_value=0.03)

    Two-parameter structural nested mean model via closed-form

    >>> snm = GEstimationSNM(df, exposure='art', outcome='cd4_wk45')
    >>> snm.exposure_model('male + age0 + age_rs1 + age_rs2 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')
    >>> snm.structural_nested_model(model='art + art:male')
    >>> snm.fit()

    Two-parameter structural nested mean model via grid-search and starting values

    >>> snm = GEstimationSNM(df, exposure='art', outcome='cd4_wk45')
    >>> snm.exposure_model('male + age0 + age_rs1 + age_rs2 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')
    >>> snm.structural_nested_model(model='art + art:male')
    >>> snm.fit(solver='search', starting_value=[-0.05, 0.0])

    References
    ----------
    Naimi AI, Cole SR, Kennedy EH. (2017). An introduction to g methods. International journal of epidemiology,
    46(2), 756-762.

    Robins JM. (2000). Marginal structural models versus structural nested models as tools for causal inference.
    In Statistical models in epidemiology, the environment, and clinical trials (pp. 95-133). Springer, New York, NY.

    Vansteelandt S, Joffe M. (2014). Structural nested models and G-estimation: the partially realized promise.
    Statistical Science, 29(4), 707-731.

    Wallace MP, Moodie EE, Stephens DA. (2017). An R package for G-estimation of structural nested mean models.
    Epidemiology, 28(2), e18-e20.

    Scharfstein DO, Rotnitzky A, Robins JM. (1999). Adjusting for nonignorable drop-out using semiparametric
    nonresponse models. Journal of the American Statistical Association, 94(448), 1096-1120.
    """
    def __init__(self, df, exposure, outcome, weights=None):
        self.exposure = exposure
        self.outcome = outcome
        self._missing_indicator = '__missing_indicator__'
        self.df, self._miss_flag, continuous = check_input_data(data=df,
                                                                exposure=exposure,
                                                                outcome=outcome,
                                                                estimator="GEstimationSNM",
                                                                drop_censoring=False,
                                                                drop_missing=True,
                                                                binary_exposure_only=True)

        self.psi = None
        self.psi_labels = None

        self._weight_ = weights
        self.ipmw = None
        self._alphas = None
        self._treatment_model = None
        self._fit_missing_ = False
        self._predicted_A = None
        self._snm_ = None
        self._print_results = True
        self._scipy_solver_obj = None

    def exposure_model(self, model, print_results=True):
        r"""Specify the treatment model to satisfy conditional exchangeability. Behind the scenes, `GestimationSNM` will
        add the necessary H(psi) terms. The only variables that need to be specified are the set of L's to satisfy
        conditional exchangeability.

        .. math::

            logit(\Pr(A=1| L)) = alpha + alpha L

        Parameters
        ----------
        model : str
            Independent variables to predict the exposure. For example, 'var1 + var2 + var3'
        print_results : bool, optional
            Whether to print the fitted model results. Default is True (prints results). This only applies to the
            closed-form solution. The grid-search procedure fits lots of regression models and these results are
            not printed

        Notes
        -----
        H(psi) terms are only necessary for the grid-search solution to g-estimation. For the closed form, we directly
        generate predicted values of A from this model. As a result, the `exposure_model()` function is agnostic to
        the estimation approach. It only requires specifying the sufficient adjustment set via patsy format
        """
        self._treatment_model = model
        self._print_results = print_results

    def structural_nested_model(self, model):
        r"""Specify the structural nested mean model to fit. The structural nested model should include the treatment of
        interest, as well as any interactions with L's that are necessary. G-estimation assumes that this model is
        correctly specified and ALL interactions with confounders are included in this model. One way to ensure
        this assumption is to saturate the structural nested mean model (or allow for as much flexibility as possible).

        The structural nested mean model takes the following form

        .. math::

            E[Y^a |A=a, V] - E[Y^a|A=a, V] = \psi a + \psi a*V

        Parameters
        ----------
        model : str
            Structural nested mean model to estimate. Model should include treatment and relevant treatment-confounder
            interaction terms as needed. Interactions should be indicated via patsy magic.
            For example, 'A + A:V + A:C'
        """
        if self.exposure not in model:
            warnings.warn("It looks like '" + self.exposure + "' is not included in the structural nested model.")

        self._snm_ = model

    def missing_model(self, model_denominator, model_numerator=None, stabilized=True, bound=False, print_results=True):
        """Estimation of Pr(M=0|A=a,L), which is the missing data mechanism for the outcome. The corresponding
        observation probabilities are used to account for informative censoring by observed variables. The missing_model
        only accounts for missing outcome data.

        The inverse probability weights calculated by this function account for informative censoring (missing data on
        the outcome) by observed variables.

        Note
        ----
        The treatment variable should be included in the model

        Parameters
        ----------
        model_denominator: str
            String listing variables predicting missingness of outcomes via `patsy` syntax. For example, `
            'var1 + var2 + var3'. This is for the predicted probabilities of the denominator
        model_numerator : str, optional
            Optional string listing variables to predict the exposure, separated by +. Only used to calculate the
            numerator. Default (None) calculates the probability of censoring by treatment only. In general this is
            recommended. If assessing effect modifcation, this variable should be included in the numerator as well.
            Argument is only used when calculating stabilized weights
        stabilized : bool, optional
            Whether to use stabilized inverse probability of censoring weights
        bound : float, list, optional
            Value between 0,1 to truncate predicted probabilities. Helps to avoid near positivity violations.
            Specifying this argument can improve finite sample performance for random positivity violations. However,
            inference becomes limited to the restricted population. Default is False, meaning no truncation of
            predicted probabilities occurs. Providing a single float assumes symmetric trunctation. A collection of
            floats can be provided for asymmetric trunctation
        print_results: bool, optional
        """
        # Error if no missing outcome data
        if not self._miss_flag:
            raise ValueError("No missing outcome data is present in the data set")

        # Warning if exposure is not included in the missingness of outcome model
        if self.exposure not in model_denominator:
            warnings.warn("For the specified missing outcome model, the exposure variable should be included in the "
                          "model", UserWarning)

        miss_model = self._missing_indicator + ' ~ ' + model_denominator
        fitmodel = propensity_score(self.df, miss_model, print_results=print_results)

        if stabilized:
            if model_numerator is None:
                mnum = self.exposure
            else:
                mnum = model_numerator
            numerator_model = propensity_score(self.df, self._missing_indicator + ' ~ ' + mnum,
                                               weights=self._weight_,
                                               print_results=print_results)
            n = numerator_model.predict(self.df)
        else:
            n = 1

        if bound:  # Bounding predicted probabilities if requested
            d = probability_bounds(fitmodel.predict(self.df), bounds=bound)
        else:
            d = fitmodel.predict(self.df)

        self.ipmw = np.where(self.df[self._missing_indicator] == 1, n / d, np.nan)
        self._fit_missing_ = True

    def fit(self, solver='closed', starting_value=None, alpha_value=0, tolerance=1e-7, verbose_solver=False,
            maxiter=500):
        """Using the treatment model and the format of the structural nested mean model, the solutions for psi are
        calculated.

        Parameters
        ----------
        solver : str, optional
            Method to use to solve for the values of psi. Default is 'closed' which uses the closed form solution,
            which is computationally less intensive.
            By specifying 'search', SciPy's optimize class is used to solve for the values of psi. One advantage of
            this approach is a sensitivity analysis approach.
        starting_value : list, array, container
            If using the 'search ' method, the initial guesses for psi can be specified. By default, no starting values
            of psi are used.
        alpha_value : float, list, array, container
            G-estimation allows for some easy and interesting sensitivity analyses. Specifically, we can imagine that
            an unmeasured confounder exists. This unmeasured confounder would mean our minimized alpha would not be
            zero anymore. We can change the alpha_value parameter to minimize alpha to different numbers. This allows
            us to see how unmeasured confounders may influence our results. This option is only available for the
            solver='search' option. See the online guide for more information on this parameter.
        verbose_solver : optional, bool
            For this grid-search procedure, setting this option to true print results for each iteration of the SciPy
            optimization procedure. Each iteration prints the psi values and the correspond alpha values from the
            fitted model
        maxiter : optional, int
            Maximum number of iterations to perform. If the maximum number of iterations is hit, the optimization
            procedure will stop and SciPy will say the convergence failed
        """
        if self._snm_ is None or self._treatment_model is None:
            raise ValueError("The exposure_model() and structural_nested_model() must be specified before the fit "
                             "procedure")

        if self._miss_flag and not self._fit_missing_:
            warnings.warn("All missing outcome data is assumed to be missing completely at random. To relax this "
                          "assumption to outcome data is missing at random please use the `missing_model()` "
                          "function", UserWarning)

        # Assigning label for weights
        df = self.df.copy()
        if self.ipmw is None:
            if self._weight_ is None:
                weight_col = None
            else:
                weight_col = self._weight_
        else:
            weight_col = '_w_'
            if self._weight_ is None:
                df['_w_'] = self.ipmw
            else:
                df['_w_'] = self.ipmw * df[self._weight_]

        # Pulling out data set up for SNM via patsy
        df = df.dropna()
        snm = patsy.dmatrix(self._snm_ + ' - 1', df, return_type='dataframe')
        self.psi_labels = snm.columns.values.tolist()  # Grabs labels for the solved psi values

        if solver == 'closed':
            # Pulling array of outcomes with the interaction terms (copy and rename column to get right interactions)
            yf = df.copy().drop(columns=[self.exposure])
            yf = yf.rename(columns={self.outcome: self.exposure})
            y_vals = patsy.dmatrix(self._snm_ + ' - 1', yf, return_type='dataframe')

            # Solving for the array of Psi values
            self.psi = self._closed_form_solver_(treat=self.exposure,
                                                 model=self.exposure + ' ~ ' + self._treatment_model,
                                                 df=df,
                                                 snm_matrix=snm, y_matrix=y_vals,
                                                 weights=weight_col, print_results=self._print_results)

        elif solver == 'search':
            # Adding other potential SNM variables to the input data
            sf = pd.concat([df, snm.drop(columns=[self.exposure])], axis=1)

            # Resolving if not initial parameters
            if starting_value is None:
                starting_value = [0.0] * len(self.psi_labels)

            # Passing to optimization procedure
            self._scipy_solver_obj = self._grid_search_(data_set=sf,
                                                        treatment=self.exposure, outcome=self.outcome,
                                                        weights=weight_col,
                                                        model=self._treatment_model, snm_terms=self.psi_labels,
                                                        start_vals=starting_value, alpha_shift=np.array(alpha_value),
                                                        tolerance=tolerance, verbose_solver=verbose_solver,
                                                        max_iter=maxiter)
            self._alphas = np.array(alpha_value)
            self.psi = self._scipy_solver_obj.x

        else:
            raise ValueError("`solver` must be specified as either 'closed' or as 'search'")

    def summary(self, decimal=3):
        """Summary of results
        """
        snm_form = ''
        is_first = False
        for l in self.psi_labels:
            if not is_first:
                is_first = True
                snm_form += 'psi*' + l
            else:
                snm_form += ' + psi*' + l

        print('======================================================================')
        print('           G-estimation of Structural Nested Mean Model               ')
        print('======================================================================')
        fmt = 'Treatment:        {:<22} No. Observations:     {:<10}'
        print(fmt.format(self.exposure, self.df.shape[0]))
        fmt = 'Outcome:          {:<22} No. Missing Outcome:  {:<10}'
        print(fmt.format(self.outcome, np.sum(self.df[self.outcome].isnull())))

        fmt = 'Missing model:    {:<15}'
        if self._fit_missing_:
            m = 'Logistic'
        else:
            m = 'None'

        print(fmt.format(m))

        # Printing scipy optimization if possible
        if self._scipy_solver_obj is not None:
            self._print_scipy_results(self._scipy_solver_obj, self._alphas)
        else:
            self._print_closed_results()

        print('----------------------------------------------------------------------')
        print('SNM:     ' + snm_form)
        print('----------------------------------------------------------------------')

        fmt = '{:<25} {:<30}'
        for p, pl in zip(self.psi, self.psi_labels):  # Printing all psi's and their labels
            print(fmt.format(pl, np.round(p, decimals=decimal)))
            # print(pl+':  ', np.round(p, decimals=decimal))

        print('======================================================================')

    @staticmethod
    def _grid_search_(data_set, treatment, outcome, model, snm_terms, start_vals,
                      alpha_shift, tolerance, verbose_solver, max_iter, weights):
        """Background function to perform the optimization procedure for psi
        """
        # Creating function for scipy to optimize
        def function_to_optimize(data, psi, snm_terms, y, a, pi_model, alpha_shift, weights):
            # loop through all psi values to calculate the corresponding H(psi) value based on covariate pattern
            snm = data[y] - data[snm_terms].mul(psi, axis='columns').sum(axis=1)
            data['H_psi'] = snm

            # Creating new terms to add to model
            h_terms_list = [w.replace(treatment, 'H_psi') for w in snm_terms]
            h_terms = ''
            for h in h_terms_list:
                h_terms += ' + ' + h

            # Estimating the necessary model
            fm = propensity_score(df=data, model=a + ' ~ ' + pi_model + h_terms,
                                  weights=weights, print_results=False)

            # Pulling elements from fitted model
            alpha = fm.params[h_terms_list] - alpha_shift  # Estimated alphas with the shift
            if verbose_solver:
                print('Psi:  ', np.array(psi))
                print('Alpha:', np.array(alpha))

            return np.abs(np.array(alpha)), psi

        def return_abs_alpha(psi):
            result = function_to_optimize(psi=psi, data=data_set,
                                          snm_terms=snm_terms,
                                          y=outcome, a=treatment,
                                          pi_model=model, alpha_shift=alpha_shift,
                                          weights=weights)
            return np.sum(result[0])

        return scipy.optimize.minimize(fun=return_abs_alpha, x0=start_vals,
                                       method='Nelder-Mead',
                                       tol=tolerance,
                                       options={'maxiter': max_iter,
                                                'disp': verbose_solver})

    @staticmethod
    def _closed_form_solver_(treat, model, df, snm_matrix, y_matrix, weights, print_results):
        """Background function to calculate the closed form solution for psi
        """
        # Calculate predictions
        fm = propensity_score(df=df, model=model, weights=weights, print_results=print_results)
        pred_treat = fm.predict(df)

        diff = df[treat] - pred_treat
        if weights is not None:
            diff = diff * df[weights]

        # D-dimensional psi-matrix
        lhm = np.dot(snm_matrix.mul(diff, axis=0).transpose(), snm_matrix)  # Dot product to produce D-by-D matrix

        # Array of outcomes
        y_matrix = y_matrix.mul(diff, axis=0)
        rha = y_matrix.sum()

        # Solving matrix and array for psi values
        psi_array = np.linalg.solve(lhm, rha)
        return psi_array

    @staticmethod
    def _print_scipy_results(optimized_function, alpha_values):
        """Background print function to the scipy optimized results results
        """
        fmt = 'Method:           {:<24} No. Iterations:   {:<10}'
        print(fmt.format('Nelder-Mead', optimized_function.nit))
        fmt = 'Alpha values:     {:<24} Optimized:        {:<10}'
        print(fmt.format(np.str(alpha_values), str(optimized_function.success)))

    @staticmethod
    def _print_closed_results():
        """Background print function to print the closed-form info
        """
        print('Method:           Closed-form')



import copy
import warnings
import patsy
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import statsmodels.api as sm
import statsmodels.formula.api as smf
from statsmodels.stats.weightstats import DescrStatsW
from scipy.stats import norm

from zepid.calc import probability_bounds
from zepid.causal.utils import (check_input_data, propensity_score, plot_kde, plot_love, iptw_calculator,
                                standardized_mean_differences, positivity,
                                plot_kde_accuracy, outcome_accuracy, aipw_calculator,
                                exposure_machine_learner, outcome_machine_learner, missing_machine_learner)


class AIPTW:
    r"""Augmented inverse probability of treatment weight estimator. This implementation calculates AIPTW for a
    time-fixed exposure and a single time-point outcome. `AIPTW` supports correcting for informative censoring (missing
    outcome data) through inverse probability of censoring/missingness weights.

    AIPTW is a doubly robust estimator, with a desirable property. Both of the the g-formula and IPTW require that
    our parametric regression models are correctly specified. Instead, AIPTW allows us to have two 'chances' at getting
    the model correct. If either our outcome-model or treatment-model is correctly specified, then our estimate
    will be unbiased. This property does not hold for the variance (i.e. the variance will not be doubly robust)

    The augment-inverse probability weight estimator is calculated from the following formula

    .. math::

        \widehat{DR}(a) = \frac{YA}{\widehat{\Pr}(A=a|L)} - \frac{\hat{Y}^a*(A-\widehat{\Pr}(A=a|L)}{
        \widehat{\Pr}(A=a|L)}

    The risk difference and risk ratio are calculated using the following formulas, respectively

    .. math::

        \widehat{RD} = \widehat{DR}(a=1) - \widehat{DR}(a=0)

    .. math::

        \widehat{RR} = \frac{\widehat{DR}(a=1)}{\widehat{DR}(a=0)}

    Confidence intervals for the risk difference come from the influence curve. Confidence intervals for the risk ratio
    are less straight-forward. To get confidence intervals for the risk ratio, a bootstrap procedure should be used.

    Parameters
    ----------
    df : DataFrame
        Pandas DataFrame object containing all variables of interest
    exposure : str
        Column name of the exposure variable. Currently only binary is supported
    outcome : str
        Column name of the outcome variable. Currently only binary is supported
    weights : str, optional
        Column name of weights. Weights allow for items like sampling weights to be used to estimate effects
    alpha : float, optional
        Alpha for confidence interval level. Default is 0.05, returning the 95% CL

    Examples
    --------
    Set up the environment and the data set

    >>> from zepid import load_sample_data, spline
    >>> from zepid.causal.doublyrobust import AIPTW
    >>> df = load_sample_data(timevary=False).drop(columns=['cd4_wk45'])
    >>> df[['cd4_rs1','cd4_rs2']] = spline(df,'cd40',n_knots=3,term=2,restricted=True)
    >>> df[['age_rs1','age_rs2']] = spline(df,'age0',n_knots=3,term=2,restricted=True)

    Estimate the base AIPTW model

    >>> aipw = AIPTW(df, exposure='art', outcome='dead')
    >>> aipw.exposure_model('male + age0 + age_rs1 + age_rs2 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')
    >>> aipw.outcome_model('art + male + age0 + age_rs1 + age_rs2 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')
    >>> aipw.fit()
    >>> aipw.summary()

    Estimate AIPTW accounting for missing outcome data

    >>> aipw = AIPTW(df, exposure='art', outcome='dead')
    >>> aipw.exposure_model('male + age0 + age_rs1 + age_rs2 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')
    >>> aipw.missing_model('art + male + age0 + age_rs1 + age_rs2 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')
    >>> aipw.outcome_model('art + male + age0 + age_rs1 + age_rs2 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')
    >>> aipw.fit()
    >>> aipw.summary()

    AIPTW for continuous outcomes

    >>> df = load_sample_data(timevary=False).drop(columns=['dead'])
    >>> df[['cd4_rs1','cd4_rs2']] = spline(df,'cd40',n_knots=3,term=2,restricted=True)
    >>> df[['age_rs1','age_rs2']] = spline(df,'age0',n_knots=3,term=2,restricted=True)

    >>> aipw = AIPTW(df, exposure='art', outcome='cd4_wk45')
    >>> aipw.exposure_model('male + age0 + age_rs1 + age_rs2 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')
    >>> aipw.missing_model('art + male + age0 + age_rs1 + age_rs2 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')
    >>> aipw.outcome_model('art + male + age0 + age_rs1 + age_rs2 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')
    >>> aipw.fit()
    >>> aipw.summary()

    >>> aipw = AIPTW(df, exposure='art', outcome='cd4_wk45')
    >>> ymodel = 'art + male + age0 + age_rs1 + age_rs2 + cd40 + cd4_rs1 + cd4_rs2 + dvl0'
    >>> aipw.exposure_model('male + age0 + age_rs1 + age_rs2 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')
    >>> aipw.missing_model(ymodel)
    >>> aipw.outcome_model(ymodel, continuous_distribution='poisson')
    >>> aipw.fit()
    >>> aipw.summary()

    References
    ----------
    Funk MJ, Westreich D, Wiesen C, Stürmer T, Brookhart MA, & Davidian M. (2011). Doubly robust
    estimation of causal effects. American Journal of Epidemiology, 173(7), 761-767.

    Lunceford JK, Davidian M. (2004). Stratification and weighting via the propensity score in estimation of causal
    treatment effects: a comparative study. Statistics in medicine, 23(19), 2937-2960.
    """

    def __init__(self, df, exposure, outcome, weights=None, alpha=0.05):
        self.exposure = exposure
        self.outcome = outcome
        self._missing_indicator = '__missing_indicator__'
        self.df, self._miss_flag, self._continuous_outcome = check_input_data(data=df,
                                                                              exposure=exposure,
                                                                              outcome=outcome,
                                                                              estimator="AIPTW",
                                                                              drop_censoring=False,
                                                                              drop_missing=True,
                                                                              binary_exposure_only=True)

        self._weight_ = weights
        self.alpha = alpha

        self.risk_difference = None
        self.risk_ratio = None
        self.risk_difference_ci = None
        self.risk_ratio_ci = None
        self.risk_difference_se = None
        self.risk_ratio_se = None

        self.average_treatment_effect = None
        self.average_treatment_effect_ci = None
        self.average_treatment_effect_se = None

        self._continuous_type = None
        self._fit_exposure_ = False
        self._exp_model_custom = False
        self._fit_outcome_ = False
        self._out_model_custom = False
        self._fit_missing_ = False
        self._miss_model_custom = False
        self._exp_model = None
        self._out_model = None
        self._predicted_y_ = None

    def exposure_model(self, model, custom_model=None, bound=False, print_results=True):
        r"""Specify the propensity score / inverse probability weight model. Model used to predict the exposure via a
        logistic regression model. This model estimates

        .. math::

            \widehat{\Pr}(A=1|L) = logit^{-1}(\widehat{\beta_0} + \widehat{\beta} L)

        Parameters
        ----------
        model : str
            Independent variables to predict the exposure. For example, 'var1 + var2 + var3'
        custom_model : optional
            Input for a custom model that is used in place of the logit model (default). The model must have the
            "fit()" and  "predict()" attributes. SciKit-Learn style models supported as custom models. In the
            background, AIPTW will fit the custom model and generate the predicted probablities
        bound : float, list, optional
            Value between 0,1 to truncate predicted probabilities. Helps to avoid near positivity violations.
            Specifying this argument can improve finite sample performance for random positivity violations. However,
            truncating weights leads to additional confounding. Default is False, meaning no truncation of
            predicted probabilities occurs. Providing a single float assumes symmetric trunctation, where values below
            or above the threshold are set to the threshold value. Alternatively a list of floats can be provided for
            asymmetric trunctation, with the first value being the lower bound and the second being the upper bound
        print_results : bool, optional
            Whether to print the fitted model results. Default is True (prints results)
        """
        self.__mweight = model
        self._exp_model = self.exposure + ' ~ ' + model

        if custom_model is None:
            d, n, iptw = iptw_calculator(df=self.df, treatment=self.exposure, model_denom=model, model_numer='1',
                                         weight=self._weight_, stabilized=False, standardize='population',
                                         bound=None, print_results=print_results)
        else:
            self._exp_model_custom = True
            data = patsy.dmatrix(model + ' - 1', self.df)
            d = exposure_machine_learner(xdata=np.asarray(data),
                                         ydata=np.asarray(self.df[self.exposure]),
                                         ml_model=copy.deepcopy(custom_model),
                                         print_results=print_results)

        g1w = d
        g0w = 1 - d

        # Applying bounds AFTER extracting g1 and g0
        if bound:
            g1w = probability_bounds(g1w, bounds=bound)
            g0w = probability_bounds(g0w, bounds=bound)

        self.df['_g1_'] = g1w
        self.df['_g0_'] = g0w
        self._fit_exposure_ = True

    def missing_model(self, model, custom_model=None, bound=False, print_results=True):
        r"""Estimation of Pr(M=0|A,L), which is the missing data mechanism for the outcome. Predicted probabilities are
        used to create inverse probability of censoring weights to account for informative missing data on the outcome.

        Missing weights take the following form

        .. math::
            \frac{1}{\Pr(C=0|A=a, L)}

        Weights are calculated for both A=1 and A=0

        Note
        ----
        The treatment variable should be included in the model

        Parameters
        ----------
        model : str
            Independent variables to predict the exposure. Example) 'var1 + var2 + var3'. The treatment must be
            included for the missing data model
        custom_model : optional
            Input for a custom model that is used in place of the logit model (default). The model must have the
            "fit()" and  "predict()" attributes. SciKit-Learn style models are supported as custom models. In the
            background, AIPTW will fit the custom model and generate the predicted probablities
        bound : float, list, optional
            Value between 0,1 to truncate predicted probabilities. Helps to avoid near positivity violations.
            Specifying this argument can improve finite sample performance for random positivity violations. However,
            truncating weights leads to additional confounding. Default is False, meaning no truncation of
            predicted probabilities occurs. Providing a single float assumes symmetric trunctation, where values below
            or above the threshold are set to the threshold value. Alternatively a list of floats can be provided for
            asymmetric trunctation, with the first value being the lower bound and the second being the upper bound
        print_results : bool, optional
            Whether to print the fitted model results. Default is True (prints results)
        """
        # Error if no missing outcome data
        if not self._miss_flag:
            raise ValueError("No missing outcome data is present in the data set")

        # Warning if exposure is not included in the missingness of outcome model
        if self.exposure not in model:
            warnings.warn("For the specified missing outcome model, the exposure variable should be included in the "
                          "model", UserWarning)

        # Warning if exposure is not included in the missingness of outcome model
        if self.exposure not in model:
            warnings.warn("For the specified missing outcome model, the exposure variable should be included in the "
                          "model", UserWarning)

        self._miss_model = self._missing_indicator + ' ~ ' + model
        fitmodel = propensity_score(self.df, self._miss_model, print_results=print_results)

        if custom_model is None:  # Logistic Regression model for predictions
            dfx = self.df.copy()
            dfx[self.exposure] = 1
            m1w = np.where(self.df[self._missing_indicator] == 1, fitmodel.predict(dfx), np.nan)
            dfx = self.df.copy()
            dfx[self.exposure] = 0
            m0w = np.where(self.df[self._missing_indicator] == 1, fitmodel.predict(dfx), np.nan)
        else:  # User-Specified model
            self._miss_model_custom = True
            data = patsy.dmatrix(model + ' - 1', self.df)
            dfx = self.df.copy()
            dfx[self.exposure] = 1
            adata = patsy.dmatrix(model + ' - 1', dfx)
            dfx = self.df.copy()
            dfx[self.exposure] = 0
            ndata = patsy.dmatrix(model + ' - 1', dfx)

            m1w, m0w = missing_machine_learner(xdata=np.array(data),
                                               mdata=self.df[self._missing_indicator],
                                               all_a=adata, none_a=ndata,
                                               ml_model=copy.deepcopy(custom_model),
                                               print_results=print_results)

        # If bounds are requested
        if bound:
            m1w = probability_bounds(m1w, bounds=bound)
            m0w = probability_bounds(m0w, bounds=bound)

        self.df['_ipmw_a1_'] = m1w
        self.df['_ipmw_a0_'] = m0w

        self._fit_missing_ = True

    def outcome_model(self, model, custom_model=None, continuous_distribution='gaussian', print_results=True):
        r"""Specify the outcome model. Model used to predict the outcome via a regression model. For binary outcome
        data, a logistic regression model is used. For continuous outcomes, either linear or Poisson regression are
        available.

        .. math::

            \widehat{\Pr}(Y|A,L) = logit^{-1}(\widehat{\beta_0} + \widehat{\beta_1} A + \widehat{\beta} L)

        Parameters
        ----------
        model : str
            Independent variables to predict the outcome. For example, 'var1 + var2 + var3 + var4'
        custom_model : optional
            Input for a custom model that is used in place of the logit model (default). The model must have the
            "fit()" and  "predict()" attributes. SciKit-Learn style models are supported as custom models. In the
            background, TMLE will fit the custom model and generate the predicted values
        continuous_distribution : str, optional
            Distribution to use for continuous outcomes. Options are 'gaussian' for normal distributions and 'poisson'
            for Poisson distributions
        print_results : bool, optional
            Whether to print the fitted model results. Default is True (prints results)
        """
        if self.exposure not in model:
            warnings.warn("It looks like '" + self.exposure + "' is not included in the outcome model.")

        if self._miss_flag:
            cc = self.df.copy().dropna()
        else:
            cc = self.df.copy()

        self._out_model = self.outcome + ' ~ ' + model

        if custom_model is None:
            if self._continuous_outcome:
                self._continuous_type = continuous_distribution
                if (continuous_distribution == 'gaussian') or (continuous_distribution == 'normal'):
                    f = sm.families.family.Gaussian()
                elif continuous_distribution == 'poisson':
                    f = sm.families.family.Poisson()
                else:
                    raise ValueError("Only 'gaussian' and 'poisson' distributions are supported")
            else:
                f = sm.families.family.Binomial()

            if self._weight_ is None:
                log = smf.glm(self._out_model, cc, family=f).fit()
            else:
                log = smf.glm(self._out_model, cc, freq_weights=cc[self._weight_], family=f).fit()

            if print_results:
                print('==============================================================================')
                print('Outcome Model')
                print(log.summary())
                print('==============================================================================')

            # Predicting under treatment strategies
            dfx = self.df.copy()
            dfx[self.exposure] = 1
            qa1w = log.predict(dfx)
            dfx = self.df.copy()
            dfx[self.exposure] = 0
            qa0w = log.predict(dfx)

        else:
            self._out_model_custom = True
            data = patsy.dmatrix(model + ' - 1', cc)

            dfx = self.df.copy()
            dfx[self.exposure] = 1
            adata = patsy.dmatrix(model + ' - 1', dfx)
            dfx = self.df.copy()
            dfx[self.exposure] = 0
            ndata = patsy.dmatrix(model + ' - 1', dfx)

            qa1w, qa0w = outcome_machine_learner(xdata=np.asarray(data),
                                                 ydata=np.asarray(cc[self.outcome]),
                                                 all_a=adata, none_a=ndata,
                                                 ml_model=copy.deepcopy(custom_model),
                                                 continuous=self._continuous_outcome,
                                                 print_results=print_results)

        self.df['_pY1_'] = qa1w
        self.df['_pY0_'] = qa0w
        self._predicted_y_ = qa1w * self.df[self.exposure] + qa0w * (1 - self.df[self.exposure])
        self._fit_outcome_ = True

    def fit(self):
        """Calculate the augmented inverse probability weights and effect measures from the predicted exposure
        probabilities and predicted outcome values.

        Note
        ----
        Exposure and outcome models must be specified prior to `fit()`

        Returns
        -------
        For binary outcomes, gains `risk_difference`, `risk_difference_ci`, and `risk_ratio` attributes. For continuous
        outcomes, gains `average_treatment_effect` and `average_treatment_effect_ci` attributes
        """
        if (self._fit_exposure_ is False) or (self._fit_outcome_ is False):
            raise ValueError('The exposure and outcome models must be specified before the doubly robust estimate can '
                             'be generated')

        if self._miss_flag and not self._fit_missing_:
            warnings.warn("All missing outcome data is assumed to be missing completely at random. To relax this "
                          "assumption to outcome data is missing at random please use the `missing_model()` "
                          "function", UserWarning)

        # Doubly robust estimator under all treated
        a_obs = np.asarray(self.df[self.exposure])
        y_obs = np.asarray(self.df[self.outcome])
        py_a1 = np.asarray(self.df['_pY1_'])
        py_a0 = np.asarray(self.df['_pY0_'])

        if self._fit_missing_:
            ps_g1 = np.asarray(self.df['_g1_'] * self.df['_ipmw_a1_'])
            ps_g0 = np.asarray(self.df['_g0_'] * self.df['_ipmw_a0_'])
        else:
            ps_g1 = np.asarray(self.df['_g1_'])
            ps_g0 = np.asarray(self.df['_g0_'])

        if self._weight_ is None:
            w = None
        else:
            w = self.df[self._weight_]
        diff_est, diff_var = aipw_calculator(y=y_obs, a=a_obs,
                                             py_a=py_a1, py_n=py_a0,
                                             pa1=ps_g1, pa0=ps_g0,
                                             difference=True, weights=w,
                                             splits=None, continuous=self._continuous_outcome)

        # Generating estimates for the risk difference and risk ratio
        zalpha = norm.ppf(1 - self.alpha / 2, loc=0, scale=1)

        if self._continuous_outcome:
            self.average_treatment_effect = diff_est
            self.average_treatment_effect_se = np.sqrt(diff_var)
            self.average_treatment_effect_ci = [self.average_treatment_effect - zalpha * np.sqrt(diff_var),
                                                self.average_treatment_effect + zalpha * np.sqrt(diff_var)]

        else:
            self.risk_difference = diff_est
            self.risk_difference_se = np.sqrt(diff_var)
            self.risk_difference_ci = [self.risk_difference - zalpha * np.sqrt(diff_var),
                                       self.risk_difference + zalpha * np.sqrt(diff_var)]

            rr, ln_rr_var = aipw_calculator(y=y_obs, a=a_obs,
                                            py_a=py_a1, py_n=py_a0,
                                            pa1=ps_g1, pa0=ps_g0,
                                            difference=False, weights=w,
                                            splits=None, continuous=False)
            self.risk_ratio = rr
            self.risk_ratio_se = np.sqrt(ln_rr_var)
            self.risk_ratio_ci = (np.exp(np.log(rr) - zalpha * self.risk_ratio_se),
                                  np.exp(np.log(rr) + zalpha * self.risk_ratio_se))

    def summary(self, decimal=3):
        """Prints a summary of the results for the doubly robust estimator. Confidence intervals are only available for
        the risk difference currently. For risk ratio confidence intervals, the user will need to manually conduct a
        boostrapping procedure.

        Parameters
        ----------
        decimal : int, optional
            Number of decimal places to display in the result
        """
        if (self._fit_exposure_ is False) or (self._fit_exposure_ is False):
            raise ValueError('The exposure and outcome models must be specified before the double robust estimate can '
                             'be generated')

        print('======================================================================')
        print('          Augmented Inverse Probability of Treatment Weights          ')
        print('======================================================================')
        fmt = 'Treatment:        {:<15} No. Observations:     {:<20}'
        print(fmt.format(self.exposure, self.df.shape[0]))

        fmt = 'Outcome:          {:<15} No. Missing Outcome:  {:<20}'
        print(fmt.format(self.outcome, np.sum(self.df[self.outcome].isnull())))

        fmt = 'g-Model:          {:<15} Missing Model:        {:<20}'
        if self._exp_model_custom:
            e = 'User-specified'
        else:
            e = 'Logistic'
        if self._miss_model_custom and self._fit_missing_:
            m = 'User-specified'
        elif self._fit_missing_:
            m = 'Logistic'
        else:
            m = 'None'

        print(fmt.format(e, m))

        fmt = 'Q-Model:          {:<15}'
        if self._out_model_custom:
            y = 'User-specified'
        elif self._continuous_outcome:
            y = self._continuous_type
        else:
            y = 'Logistic'

        print(fmt.format(y))

        print('======================================================================')

        if self._continuous_outcome:
            print('Average Treatment Effect:   ', round(float(self.average_treatment_effect), decimal))
            if self._weight_ is None:
                print(str(round(100 * (1 - self.alpha), 1)) + '% two-sided CI: (' +
                      str(round(self.average_treatment_effect_ci[0], decimal)), ',',
                      str(round(self.average_treatment_effect_ci[1], decimal)) + ')')
            else:
                print(str(round(100 * (1 - self.alpha), 1)) + '% two-sided CI: -')
        else:
            print('Risk Difference:    ', round(float(self.risk_difference), decimal))
            if self._weight_ is None:
                print(str(round(100 * (1 - self.alpha), 1)) + '% two-sided CI: (' +
                      str(round(self.risk_difference_ci[0], decimal)), ',',
                      str(round(self.risk_difference_ci[1], decimal)) + ')')
            else:
                print(str(round(100 * (1 - self.alpha), 1)) + '% two-sided CI: -')
            print('----------------------------------------------------------------------')
            print('Risk Ratio:        ', round(float(self.risk_ratio), decimal))
            if self._weight_ is None:
                print(str(round(100 * (1 - self.alpha), 1)) + '% two-sided CI: (' +
                      str(round(self.risk_ratio_ci[0], decimal)), ',',
                      str(round(self.risk_ratio_ci[1], decimal)) + ')')
            else:
                print(str(round(100 * (1 - self.alpha), 1)) + '% two-sided CI: -')

        print('======================================================================')

    def run_diagnostics(self, decimal=3):
        """Run all currently implemented diagnostics for the exposure and outcome models. Each
        `run_diagnostics` provides results for all implemented diagnostics for ease of the user. For publication
        quality presentations, I recommend calling each diagnostic function individually and utilizing the optional
        parameters

        Note
        ----
        The plot presented cannot be edited. To edit the plots, call `plot_kde` or `plot_love` directly. Those
        functions return an axes object

        Parameters
        ----------
        decimal : int, optional
            Number of decimal places to display. Default is 3

        Returns
        -------
        None
        """
        if not self._fit_outcome_ or not self._fit_exposure_:
            raise ValueError("The exposure_model and outcome_model function must be ran before any diagnostics")

        # Weight diagnostics
        print('\tExposure Model Diagnostics')
        self.positivity(decimal=decimal)

        print('\n======================================================================')
        print('                  Standardized Mean Differences')
        print('======================================================================')
        print(self.standardized_mean_differences().set_index(keys='labels'))
        print('======================================================================\n')

        # Outcome accuracy diagnostics
        print('\tOutcome Model Diagnostics')
        v = self._predicted_y_ - self.df[self.outcome]
        outcome_accuracy(true=self.df[self.outcome], predicted=self._predicted_y_, decimal=decimal)

        df = self.df.copy()
        df['_ipw_'] = np.where(df[self.exposure] == 1, 1 / df['_g1_'], 1 / (1 - df['_g1_']))

        plt.figure(figsize=[8, 6])
        plt.subplot(221)
        plot_love(df=df, treatment=self.exposure, weight='_ipw_', formula=self.__mweight)
        plt.title("Love Plot")

        plt.subplot(223)
        plot_kde(df=df, treatment=self.exposure, probability='_g1_')
        plt.title("Kernel Density of Propensity Scores")

        plt.subplot(222)
        plot_kde_accuracy(values=v.dropna(), color='green')
        plt.title("Kernel Density of Accuracy")
        plt.tight_layout()
        plt.show()

    def positivity(self, decimal=3):
        """Use this to assess whether positivity is a valid assumption for the exposure model / calculated IPTW. If
        there are extreme outliers, this may indicate problems with the calculated weights. To reduce extreme weights,
        the `bound` argument can be specified in `exposure_model()`

        Parameters
        ----------
        decimal : int, optional
            Number of decimal places to display. Default is three

        Returns
        -------
        None
            Prints the positivity results to the console but does not return any objects
        """
        df = self.df.copy()
        df['_ipw_'] = np.where(df[self.exposure] == 1, 1 / df['_g1_'], 1 / (1 - df['_g1_']))

        pos = positivity(df=df, weights='_ipw_')
        print('======================================================================')
        print('                      Weight Positivity Diagnostics')
        print('======================================================================')
        print('If the mean of the weights is far from either the min or max, this may\n '
              'indicate the model is incorrect or positivity is violated')
        print('Average weight should be 2')
        print('----------------------------------------------------------------------')
        print('Mean weight:           ', round(pos[0], decimal))
        print('Standard Deviation:    ', round(pos[1], decimal))
        print('Minimum weight:        ', round(pos[2], decimal))
        print('Maximum weight:        ', round(pos[3], decimal))
        print('======================================================================\n')

    def standardized_mean_differences(self):
        """Calculates the standardized mean differences for all variables based on the inverse probability weights.

        Returns
        -------
        DataFrame
            Returns pandas DataFrame of calculated standardized mean differences. Columns are labels (variables labels),
            smd_u (unweighted standardized difference), and smd_w (weighted standardized difference)
        """
        df = self.df.copy()
        df['_ipw_'] = np.where(df[self.exposure] == 1, 1 / df['_g1_'], 1 / (1 - df['_g1_']))
        s = standardized_mean_differences(df=df, treatment=self.exposure,
                                          weight='_ipw_', formula=self.__mweight)
        return s

    def plot_kde(self, to_plot, bw_method='scott', fill=True, color='g', color_e='b', color_u='r'):
        """Generates density plots that can be used to check predictions qualitatively. Density plots can be generated
        for assess either positivity violations of the exposure model or the accuracy in predicting the outcome for
        the outcome model. The kernel density used is SciPy's Gaussian kernel. Either Scott's Rule or
        Silverman's Rule can be implemented.

        Parameters
        ------------
        to_plot : str, optional
            The plot to generate. Specifying 'exposure' returns only the density plot for treatment probabilities,
            and 'outcome' returns only the density plot for the outcome accuracy
        bw_method : str, optional
            Method used to estimate the bandwidth. Following SciPy, either 'scott' or 'silverman' are valid options
        fill : bool, optional
            Whether to color the area under the density curves. Default is true
        color : str, optional
            Color of the line/area for predicted outcomes minus observed outcomes. Default is Green
        color_e : str, optional
            Color of the line/area for the treated group. Default is Blue
        color_u : str, optional
            Color of the line/area for the treated group. Default is Red

        Returns
        ---------------
        matplotlib axes
        """
        if to_plot == 'exposure':
            ax = plot_kde(df=self.df, treatment=self.exposure, probability='_g1_',
                          bw_method=bw_method, fill=fill, color_e=color_e, color_u=color_u)
            ax.set_title("Kernel Density of Propensity Scores")

        elif to_plot == 'outcome':
            v = self._predicted_y_ - self.df[self.outcome]
            ax = plot_kde_accuracy(values=v.dropna(), bw_method=bw_method, fill=fill, color=color)
            ax.set_title("Kernel Density of Accuracy")

        else:
            raise ValueError("Please use one of the following options for `to_plot`; 'treatment', 'outcome'")

        return ax

    def plot_love(self, color_unweighted='r', color_weighted='b', shape_unweighted='o', shape_weighted='o'):
        """Generates a Love-plot to detail covariate balance based on the IPTW weights. Further details on the usage of
        this plot are available in Austin PC & Stuart EA 2015 https://onlinelibrary.wiley.com/doi/full/10.1002/sim.6607

        The Love plot generates a dashed line at standardized mean difference of 0.10. Ideally, weighted SMD are below
        this level. Below 0.20 may also be sufficient. Variables above this level may be unbalanced despite the
        weighting procedure. Different functional forms (or approaches like machine learning) may be worth considering

        Parameters
        ----------
        color_unweighted : str, optional
            Color for the unweighted standardized mean differences. Default is red
        color_weighted : str, optional
            Color for the weighted standardized mean differences. Default is blue
        shape_unweighted : str, optional
            Shape of points for the unweighted standardized mean differences. Default is circles
        shape_weighted:
            Shape of points for the weighted standardized mean differences. Default is circles

        Returns
        -------
        axes
            Matplotlib axes of the Love plot
        """
        df = self.df.copy()
        df['_ipw_'] = np.where(df[self.exposure] == 1, 1 / df['_g1_'], 1 / (1 - df['_g1_']))

        ax = plot_love(df=df, treatment=self.exposure, weight='_ipw_', formula=self.__mweight,
                       color_unweighted=color_unweighted, color_weighted=color_weighted,
                       shape_unweighted=shape_unweighted, shape_weighted=shape_weighted)
        return ax



import copy
import warnings
import numpy as np
import pandas as pd
import sklearn.model_selection as ms

from scipy.optimize import nnls
from sklearn import clone

from zepid.calc import logit, inverse_logit, probability_bounds


class SuperLearnerError(Exception):
    """Class for errors in the SuperLearner procedures. Nothing special besides directing user to specific issues are
    to the SuperLearner side
    """
    pass


class SuperLearner:
    r"""`SuperLearner` is an implementation of the super learner algorithm, which is a generalized stacking algorithm.
    Super learner is an approach to combine multiple predictive functions into a singular predictive function that has
    performance at least as good as the best candidate estimator included (asymptotically). Additionally, it should be
    noted that super learner converges at the rate with which the best candidate estimator converges.

    Briefly, super learner takes an input of candidate estimators for a function. Each of the estimators is run through
    a train-test cross-validation algorithm. From the candidate estimators, either the best overall performing
    candidate (discrete super learner) or a weighted combination of the algorithms is used as the updated predictive
    function.

    Note
    ----
    `SuperLearner` does not accept missing data. All missing data decisions have to occur prior to trying to use the
    `SuperLearner` procedure.


    `SuperLearner` accepts estimators that are of the SciKit-Learn format. Specifically, the candidate estimators must
    follow the `estimator.fit(X, y)` and `estimator.predict(X)` format. Performance has currently been checked for
    `sklearn`, `pygam`, and the estimators included in `zepid.superlearner`. Please consider opening an issue on
    GitHub if you find Python libraries that are not supported (but follow the SciKit-Learn style).

    Note
    ----
    `SuperLearner(discrete=True)` returns predictions from the candidate estimator with the greatest coefficient. In the
    case of a tie, the first candidate estimator with the greatest coefficient is used (as per `numpy.argmax` behavior).


    To compare performances easily, `SuperLearner` provides both Cross-Validated Error and the Relative Efficiency. The
    Cross-Validated Error calculation depends on the chosen loss function. For L2, the loss function is

    .. math::

        \frac{1}{n} \sum_i (Y_i - \widehat{Y}_i)^2

    For the negative-log-likelihood loss function,

    .. math::

        \frac{1}{n} \sum_i Y_i \times \ln(\widehat{Y}_i) + (1-Y_i) \times \ln(1 - \widehat{Y}_i)

    Relative efficiency is the Cross-Validated Error for the candidate estimator divided by the Cross-Validated Error
    for the chosen super learner.

    Parameters
    ----------
    estimators : list, array
        Candidate estimators. Must follow sklearn style and not be fit yet
    estimator_labels : list, array
        Labels for the candidate estimators being included
    folds : int, optional
        Number of folds to use during the cross-validation procedure. It is recommended to be between 10-20. The default
        value is 10-fold cross-validation.
    loss_function : str, optional
        Loss function to use. Options include: L2, NLogLik. L2 should be used for continuous outcomes and NLogLik for
        binary outcomes
    solver : str, optional
        Optimization algorithm to use to determine the super learner weights. Currently only Non-Negative Least Squares
        is available.
    bounds : float, collection, optional
        Bounding to use for probability. The bounding prevents values of exactly 0 or 1, which will break the loss
        function evaluation. Default is 1e-6.
    discrete : bool, optional
        Whether to use only the estimator with the greatest weight (discrete super learner). Default is False, which
        uses the super learner including all estimators
    verbose : bool, optional
        Whether to print progress to the console as super learner is being fit. Default is False.

    Examples
    --------
    Setup the environment and data set

    >>> import statsmodels.api as sm
    >>> from sklearn.linear_model import LinearRegression, LogisticRegression
    >>> from zepid import load_sample_data
    >>> from zepid.superlearner import EmpiricalMeanSL, StepwiseSL, SuperLearner

    >>> fb = sm.families.family.Gaussian()
    >>> fc = sm.families.family.Binomial()
    >>> df = load_sample_data(False).dropna()
    >>> X = np.asarray(df[['art', 'male', 'age0']])
    >>> y = np.asarray(df['dead'])

    SuperLearner for binary outcomes

    >>> # Setting up estimators
    >>> emp = EmpiricalMeanSL()
    >>> log = LogisticRegression()
    >>> step = StepwiseSL(family=fb, selection="backward", order_interaction=1)
    >>> sl = SuperLearner(estimators=[emp, log, step], estimator_labels=["Mean", "Log", "Step"], loss_function='nloglik')
    >>> fsl = sl.fit(X, y)
    >>> fsl.summary()  # Summary of Cross-Validated Errors
    >>> fsl.predict(X)  # Generating predicted values from super learner

    SuperLearner for continuous outcomes

    >>> emp = EmpiricalMeanSL()
    >>> lin = LinearRegression()
    >>> step = StepwiseSL(family=fc, selection="backward", order_interaction=1)
    >>> sl = SuperLearner(estimators=[emp, log, step], estimator_labels=["Mean", "Lin", "Step"], loss_function='L2')
    >>> fsl = sl.fit(X, y)
    >>> fsl.summary()  # Summary of Cross-Validated Errors
    >>> fsl.predict(X)  # Generating predicted values from super learner

    Discrete Super Learner

    >>> sl = SuperLearner([emp, log, step], ["Mean", "Lin", "Step"], loss_function='L2', discrete=True)
    >>> sl.fit(X, y)

    References
    ----------
    Van der Laan MJ, Polley EC, Hubbard AE. (2007). Super learner. Statistical Applications in Genetics and Molecular
    Biology, 6(1).

    Rose S. (2013). Mortality risk score prediction in an elderly population using machine learning. American Journal
    of Epidemiology, 177(5), 443-452.
    """
    def __init__(self, estimators, estimator_labels, folds=10, loss_function="L2", solver="nnls",
                 bounds=1e-6, discrete=False, verbose=False):
        # TODO the R SuperLearner library supports an alt-NNLS, LS, and nnloglik optimization routines. Find in Python
        # Checking for errors
        if len(estimators) != len(estimator_labels):
            raise ValueError("estimators and estimator_labels must be of the same length")
        if solver.lower() not in ["nnls"]:
            raise ValueError("The solver " + str(solver) + " is not currently available. Please select one of the "
                                                           "following: NNLS")
        if loss_function.lower() not in ["l2", "nloglik"]:
            raise ValueError("The loss function " + str(loss_function) + " is not currently available. Please select "
                                                                         "one of the following: L2, NLogLik")

        # Parameters
        self.estimators = estimators
        self.labels = estimator_labels
        self.k = folds
        self.loss_function = loss_function.lower()
        self.discrete = discrete
        self.solver = solver.lower()
        self._verbose_ = verbose
        self._bounds_ = bounds

        # Storage items for results
        self.est_performance = pd.DataFrame()
        self.est_performance['estimator'] = list(estimator_labels)
        self.coefficients = None
        self.fit_estimators = []
        self._include_estimator_ = []

    def fit(self, X, y):
        """Fit SuperLearner given the variables `X` to predict `y`. These variables are directly passed to the
        candidate estimators. If there is any pre-processing to do outside of the estimator, please do so before passing
        to fit.

        Parameters
        ----------
        X : numpy.array
            Covariates to predict the target values
        y : numpy.array
            Target values to predict

        Returns
        -------
        None
        """
        # Some checks
        X = np.asarray(X)  # this line saves a lot of headaches
        y = np.asarray(y)  # this line saves a lot of headaches
        if X.shape[0] != y.shape[0]:
            raise ValueError("X and y must have the same number of observations.")
        if np.any(np.isnan(X)) or np.any(np.isnan(y)):
            raise ValueError("It looks like there is missing values in X or y. SuperLearner does not support missing "
                             "data.")
        if np.all(np.in1d(y, np.array([0, 1]))) and (self.loss_function == "l2"):
            # Allows for the algorithm to proceed (but throws warning to the user about the chosen loss function)
            warnings.warn("It looks like your `y` is binary, and the `L2` loss function should be used for "
                          "continuous outcomes", UserWarning)

        # Step 1) input data and algorithms
        n_obs = X.shape[0]  # number of observations
        n_est = len(self.estimators)  # number of candidate estimators
        cv_pred = np.full([n_obs, n_est], np.nan)  # NaN for blank results

        # Step 2) Cross-Validation Splits
        if self._verbose_:
            print("Starting cross-validation procedure...")
        current_fold = 0
        for train, test in ms.KFold(self.k, shuffle=False).split(range(n_obs)):
            current_fold += 1

            # Step 3) Test-Train for each algorithm for each fold (note that train is only a 1/folds of the hold-out)
            X_train, X_test = X[train], X[test]
            y_train, y_test = y[train], y[test]
            for est_id in range(n_est):
                if self._verbose_:
                    print("...fitting "+str(self.labels[est_id])+" fold-"+str(current_fold))
                cv_est = clone(self.estimators[est_id])
                cv_est.fit(X_train, y_train)

                # Step 4) Filling in cross-validated Y predictions
                cv_pred[test, est_id] = self._predict_(estimator=cv_est, X=X_test)

        # Step 5) Calculating Cross-Validated Error for each candidate
        self.est_performance['cv_error'] = np.nan
        for est_id in range(n_est):
            self.est_performance.loc[est_id, "cv_error"] = self._error_term_(y, cv_pred[:, est_id])

        # Step 6) Determine coefficients
        if self.solver == "nnls":
            coefs, _ = nnls(cv_pred, y)
        else:
            raise ValueError(str(self.solver) + " is not currently available")

        coefs = np.array(coefs)
        machine_limit = np.finfo(np.double).eps
        coefs[coefs < np.sqrt(machine_limit)] = 0  # dropping coefs below certain precision
        self.coefficients = coefs / np.sum(coefs)
        self.est_performance['coefs'] = self.coefficients

        # Step 7) Fit algorithms to full data
        if self._verbose_:
            print("Fitting candidate(s) to full data...")

        if self.discrete:  # Step 7.a) discrete super learner
            discrete_sl_id = np.argmax(self.coefficients)
            if self._verbose_:
                print("Discrete super learner is: " + str(self.labels[discrete_sl_id]))
            for est_id in range(n_est):
                est = clone(self.estimators[est_id])
                # Estimating only the Discrete Super Learner to save computation time
                if est_id == discrete_sl_id:
                    if self._verbose_:
                        print("...fitting " + str(self.labels[est_id]))
                    est.fit(X, y)
                    self.fit_estimators.append(est)
                    self.coefficients[est_id] = 1
                # Skipping over others but retaining the shape of input estimators
                else:
                    self.fit_estimators.append(est)
                    self.coefficients[est_id] = 0

        else:  # Step 7.b) super learner
            # No prior selection (unlike discrete super learner)
            for est_id in range(n_est):
                est = clone(self.estimators[est_id])
                # Estimating only candidate estimators with coefficient > 0
                if self.coefficients[est_id] > 0:
                    if self._verbose_:
                        print("...fitting " + str(self.labels[est_id]))
                    est.fit(X, y)
                    self.fit_estimators.append(est)
                # Skipping over candidates with coefficient == 0
                else:
                    if self._verbose_:
                        print("...skipping " + str(self.labels[est_id]))
                    self.fit_estimators.append(est)

        return self

    def predict(self, X):
        """Generate predictions using the fit SuperLearner.

        Parameters
        ----------
        X : numpy.array
            Covariates to generate predictions of y. Note that X should be in the same format as the X used during the
            fit() function

        Returns
        -------
        numpy.array of predicted values using either discrete super learner or super learner
        """
        X = np.asarray(X)  # this line saves a lot of headaches
        if self.coefficients is None:
            raise ValueError("fit() must be called before predict()")
        if np.any(np.isnan(X)):
            raise ValueError("It looks like there is missing values in X. SuperLearner does not support missing data.")

        n_obs = X.shape[0]
        n_est = len(self.estimators)
        cv_pred = np.full([n_obs, n_est], np.nan)  # NaN for blank results

        # Step 8) Predictions from SuperLearner
        for est_id in range(n_est):
            if self.coefficients[est_id] > 0:
                cv_pred[:, est_id] = self._predict_(self.fit_estimators[est_id], X)
            else:
                cv_pred[:, est_id] = 0

        # Combining predictions
        if self.loss_function == "l2":
            y_pred = np.dot(cv_pred, self.coefficients)
        if self.loss_function == 'nloglik':
            cv_pred_bound = probability_bounds(cv_pred, bounds=self._bounds_)
            logodds = logit(cv_pred_bound)
            logodds_pred = np.dot(logodds, self.coefficients)
            y_pred = inverse_logit(logodds_pred)

        return y_pred

    def summary(self):
        """Prints the summary information for the fit SuperLearner to the console.

        Returns
        -------
        None
        """
        if self.coefficients is None:
            raise ValueError("fit() must be called before summary()")
        print('======================================================================')
        print("            Super Learner Candidate Estimator Performance             ")
        print('======================================================================')
        print(self.est_performance.set_index("estimator"))
        print('======================================================================')

    def _predict_(self, estimator, X):
        """Background function to generate predictions based on the designated loss-function
        """
        if self.loss_function == 'l2':
            pred = estimator.predict(X)

        if self.loss_function == 'nloglik':
            if hasattr(estimator, "predict_proba"):
                try:  # Allows for use with PyGAM
                    pred = estimator.predict_proba(X)[:, 1]
                except IndexError:
                    pred = estimator.predict_proba(X)
                if pred.min() < 0 or pred.max() > 1:
                    raise SuperLearnerError("Probability less than zero or greater than one")
            else:
                pred = estimator.predict(X)
                if pred.min() < 0 or pred.max() > 1:
                    raise SuperLearnerError("Probability less than zero or greater than one")

        return pred

    def _error_term_(self, y_obs, y_pred):
        """Calculates the Error term based on the loss function
        """
        if self.loss_function == "l2":
            error = np.sum((y_obs - y_pred) ** 2) / y_obs.shape[0]
        if self.loss_function == 'nloglik':
            y_pred_bound = probability_bounds(v=y_pred, bounds=self._bounds_)
            error = - np.sum(y_obs*y_pred_bound + (1-y_obs)*np.log(1-y_pred_bound)) / y_obs.shape[0]
        return error



import warnings
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import statsmodels.api as sm
import statsmodels.formula.api as smf

from zepid.causal.utils import check_input_data, outcome_accuracy, plot_kde_accuracy


class TimeFixedGFormula:
    r"""G-formula for time-fixed exposure and single endpoint, also referred to as the g-computation algorithm formula.
    Uses the Snowden trick to calculate the marginal treatment under the specified exposure plan

    The g-formula can be expressed as

    .. math::

        E[Y^a] = \sum_l E[Y|A=a,L=l] \times \Pr(L=l)

    When L is continuous, the summation becomes an integral.

    Currently, `TimeFixedGFormula` only supports binary or continuous outcomes. For binary outcomes a logistic
    regression model to predict probabilities of outcomes via statsmodels. For continuous outcomes a linear regression
    or a Poisson regression model can be used to predict outcomes.

    Binary and multivariate exposures are supported. For binary exposures, a string object of the column name for
    the exposure of interest should be provided. For multivariate exposures, a list of string objects corresponding
    to disjoint indicator terms for the exposure should be provided. Multivariate exposures require the user to
    custom specify treatments when fitting the g-formula. A list of the custom treatment must be provided and be
    the same length as the number of disjoint indicator columns. See
    https://github.com/pzivich/Python-for-Epidemiologists/tree/master/3_Epidemiology_Analysis/c_causal_inference/1_time-fixed-treatments
    for examples (highly recommended)

    Key options for treatments:

        * `'all'`     -all individuals are given treatment
        * `'none'`    -no individuals are given treatment
        * custom treatments -create a custom treatment. When specifying this, the dataframe must be referred to as 'g'.
          The following is an example that selects those whose age is 30 or younger and are females:
          ``treatment="((g['age0']<=30) & (g['male']==0))``

    Note
    ----
    Custom treatments use a "magic-g" parameter. Internally, the g-formula implementation names the data set as `g`.
    Therefore, when using custom treatment specifications, the data set must be referred to as `g` when following
    the pandas selection syntax

    Parameters
    ----------
    df : DataFrame
        Pandas dataframe containing the variables of interest
    exposure : str, list
        Column name for exposure variable label or a list of disjoint indicator exposures
    outcome : str
        Column name for outcome variable
    outcome_type : str, optional
        Outcome variable type. Currently only 'binary', 'normal', and 'poisson variable types are supported
    standardize : str, optional
        Who the estimate corresponds to. Options are the entire population, the exposed, or the unexposed. See
        Sato & Matsuyama Epidemiology (2003) for details on weighting to exposed/unexposed. Weighting to the
        exposed or unexposed is also referred to as SMR weighting. Options for standardization are:
        * 'population'    :   weight to entire population
        * 'exposed'       :   weight to exposed individuals
        * 'unexposed'     :   weight to unexposed individuals
    weights : str, optional
        Column name for weights. Default is None, which assumes every observations has the same weight (i.e. 1)

    Examples
    --------
    Setting up the environment

    >>> from zepid import load_sample_data, spline
    >>> from zepid.causal.gformula import TimeFixedGFormula
    >>> df = load_sample_data(timevary=False)
    >>> df[['cd4_rs1', 'cd4_rs2']] = spline(df, 'cd40', n_knots=3, term=2, restricted=True)
    >>> df[['age_rs1', 'age_rs2']] = spline(df, 'age0', n_knots=3, term=2, restricted=True)

    G-formula with a binary treatment and outcome

    >>> g = TimeFixedGFormula(df, exposure='art', outcome='dead')
    >>> g.outcome_model(model='art + male + age0 + age_rs1 + age_rs2 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')

    >>> # Return the estimated marginal outcome under treat-all
    >>> g.fit(treatment='all')
    >>> g.marginal_outcome

    >>> # Return the estimated marginal outcome under treat-none
    >>> g.fit(treatment='all')
    >>> g.marginal_outcome

    >>> # Return the estimated marginal outcome under custom treatment (treat all females under 40)
    >>> g.fit(treatment="((g['male']==0) & (g['age0']<=40))")
    >>> g.marginal_outcome

    G-formula with a categorical treatment and binary outcome

    >>> # Creating categorical variable for CD4 count
    >>> df['cd4_1'] = np.where(((df['cd40'] >= 200) & (df['cd40'] < 400)), 1, 0)
    >>> df['cd4_2'] = np.where(df['cd40'] >= 400, 1, 0)

    >>> g = TimeFixedGFormula(df,exposure=['art_male', 'art_female'], outcome='dead', exposure_type='categorical')
    >>> g.outcome_model(model='cd4_1 + cd4_2 + art + male + age0 + age_rs1 + age_rs2 + dvl0')

    >>> # Return marginal outcome under all in reference category (CD4 < 200)
    >>> g.fit(treatment=["False", "False"])

    >>> # Return marginal outcome under all in category 1 (CD4 >= 200 & CD4 < 400)
    >>> g.fit(treatment=["True", "False"])

    >>> # Return marginal outcome under all in category 2 (CD4 > 400)
    >>> g.fit(treatment=["False", "True"])

    G-formula with binary exposure and continuous (normal-distributed) outcome

    >>> g = TimeFixedGFormula(df,exposure='art', outcome='cd4', outcome_type='normal')
    >>> g.outcome_model(model='art + male + age0 + age_rs1 + age_rs2 + dvl0  + cd40 + cd4_rs1 + cd4_rs2')

    G-formula with binary exposure and continuous (Poisson-distributed) outcome

    >>> g = TimeFixedGFormula(df,exposure='art', outcome='cd4', outcome_type='poisson')
    >>> g.outcome_model(model='art + male + age0 + age_rs1 + age_rs2 + dvl0  + cd40 + cd4_rs1 + cd4_rs2')

    G-formula with binary outcome and exposure. With a stochastic treatment/intervention

    >>> g = TimeFixedGFormula(df,exposure='art', outcome='cd4', outcome_type='poisson')
    >>> g.outcome_model(model='art + male + age0 + age_rs1 + age_rs2 + dvl0  + cd40 + cd4_rs1 + cd4_rs2')
    >>> g.fit_stochastic(p=0.75)

    G-formula with binary outcome and exposure. With a conditional stochastic treatment/intervention

    >>> g = TimeFixedGFormula(df,exposure='art', outcome='cd4')
    >>> g.outcome_model(model='art + male + age0 + age_rs1 + age_rs2 + dvl0  + cd40 + cd4_rs1 + cd4_rs2')
    >>> g.fit_stochastic(p=[0.65, 0.85], conditional=["g['male']==1", "g['male']==0"])

    References
    ----------
    JM Snowden, S Rose, and KM Mortimer. "Implementation of G-computation on a simulated
    data set: demonstration of a causal inference technique." American Journal of Epidemiology 173.7 (2011):
    731-738.

    J Ahern, KE Colson, C Margerson-Zilko, A Hubbard, & S Galea. (2016). Predicting the population
    health impacts of community interventions: the case of alcohol outlets and binge drinking. American
    Journal of Public Health, 106(11), 1938-1943.

    J Ahern, A Hubbard, & S Galea. (2009). Estimating the effects of potential public health interventions on
    population disease burden: a step-by-step illustration of causal inference methods. American Journal of
    Epidemiology, 169(9), 1140-1147.
    """
    def __init__(self, df, exposure, outcome, exposure_type='binary', outcome_type='binary', standardize='population',
                 weights=None):
        self.exposure = exposure
        self.outcome = outcome
        self._missing_indicator = '__missing_indicator__'
        self.gf, self._miss_flag, self._continuous_outcome_ = check_input_data(data=df,
                                                                               exposure=exposure,
                                                                               outcome=outcome,
                                                                               estimator="TimeFixedGFormula",
                                                                               drop_censoring=False,
                                                                               drop_missing=True,
                                                                               binary_exposure_only=False)

        if (outcome_type == 'binary') or (outcome_type == 'normal') or (outcome_type == 'poisson'):
            self.outcome_type = outcome_type
        else:
            raise ValueError('Only binary or continuous outcomes are currently supported. Please specify "binary" '
                             '"normal", or "poisson"')

        if (exposure_type == 'binary') or (exposure_type == 'continuous') or (exposure_type == 'categorical'):
            self.exposure_type = exposure_type
        else:
            raise ValueError('Only binary or continuous exposures are currently supported. Please specify "binary", '
                             '"categorical", or "continuous".')

        if standardize in ['population', 'exposed', 'unexposed']:
            self.standardize = standardize
        else:
            raise ValueError('Please specify one of the currently supported standardizations: ' +
                             'population, exposed, unexposed')

        self._weights = weights
        self._outcome_model = None
        self._predicted_y_ = None
        self.marginal_outcome = np.nan
        self.predicted_df = None

    def outcome_model(self, model, print_results=True):
        """Build the outcome regression model. This is also referred to at the Q-model in various parts of the
        literature. This must be specified before the fit function. It is encouraged to make this model as flexible as
        possible

        Parameters
        ----------
        model : str
            Variables to include in the model for predicting the outcome. Must be contained within the input
            pandas dataframe when initialized. Model form should contain the exposure, i.e. 'art + age + male'
        print_results : bool, optional
            Whether to print the logistic regression results to the terminal. Default is True
        """
        if type(self.exposure) is not list:
            if self.exposure not in model:
                warnings.warn("It looks like '" + self.exposure + "' is not included in the outcome model.")

        if self.outcome_type == 'binary':
            linkdist = sm.families.family.Binomial()
        elif self.outcome_type == 'normal':
            linkdist = sm.families.family.Gaussian()
        else:
            linkdist = sm.families.family.Poisson()

        # Modeling the outcome
        if self._weights is None:
            m = smf.glm(self.outcome + ' ~ ' + model, self.gf, family=linkdist)
            self._outcome_model = m.fit()
        else:
            m = smf.glm(self.outcome + ' ~ ' + model, self.gf, family=linkdist,
                        freq_weights=self.gf[self._weights])
            self._outcome_model = m.fit()

        # Creating predicted Y variable
        self._predicted_y_ = self._outcome_model.predict(self.gf)

        # Printing results of the model and if any observations were dropped
        if print_results:
            print('==============================================================================')
            print('Outcome Model')
            print(self._outcome_model.summary())
            print('==============================================================================')

    def fit(self, treatment, predict_missing=True):
        """Fit the parametric g-formula as specified. Binary and multivariate treatments are available. This
        implementation has three options for the binary treatment courses. For multivariate treatments, the user must
        specify custom treatment plans.

        To obtain the confidence intervals, use a bootstrap procedure

        Parameters
        ----------
        treatment : str, list
            There are three options available for treatment plans. All, none, or a custom pattern
              * `'all'`     -all individuals are given treatment
              * `'none'`    -no individuals are given treatment
              * custom  -create a custom treatment. When specifying this, the dataframe must be referred to as 'g' The
                following is an example that selects those whose age is 25 or older and are females;
                ``treatment="((g['age0']>=25) & (g['male']==0))``
        predict_missing : bool, optional
            Whether missing outcome values should be predicted via the model fit to the data with observed outcomes.
            Default is set to True, which assumes that outcome data is missing completely at random conditional on
            the variables included in the outcome model. Note this is the less restrictive assumption regarding missing
            outcome data. If set to False, missing outcome data is assumed to be missing completely at random
            (also referred to as non-informative censoring)

        Returns
        -------
        marginal_outcome
            Parameter of `marginal_outcome` is filled, which is the mean predicted outcome under the treatment strategy
        """
        if self._outcome_model is None:
            raise ValueError('Before the g-formula can be calculated, the outcome model must be specified')
        if (type(treatment) != str) and (type(treatment) != list):
            raise ValueError('Specified treatment must be a string object or a list of string objects')

        # Setting outcome as blank
        g = self.gf.copy()

        # Setting treatment (either multivariate or binary)
        if self.exposure_type == 'binary':
            if type(treatment) == list:
                raise ValueError('A binary exposure is specified. Treatment plan should be a string object')
            if treatment == 'all':
                g[self.exposure] = 1
            elif treatment == 'none':
                g[self.exposure] = 0
            else:  # custom exposure pattern
                g[self.exposure] = np.where(eval(treatment), 1, 0)

        elif self.exposure_type == 'categorical':
            if (treatment == 'all') or (treatment == 'none'):  # Check to make sure custom treatment
                raise ValueError('A multivariate exposure has been specified. A custom treatment must be '
                                 'specified by the user')
            else:
                if len(self.exposure) != len(treatment):  # Check to make sure same about of treatments specified
                    raise ValueError('The list of custom treatment conditions must be the same size as the number of '
                                     'treatments')
                for i in range(len(self.exposure)):
                    g[self.exposure[i]] = np.where(eval(treatment[i]), 1, 0)

                if np.sum(np.where(g[self.exposure].sum(axis=1) > 1, 1, 0)) > 1:
                    warnings.warn('It looks like your specified treatment strategy results in some individuals '
                                  'receiving at least two exposures. Reconsider how the custom treatments are '
                                  'specified', UserWarning)

        else:
            raise ValueError('Still working on allowing for continuous treatments...')
            # TODO fill in this part of continuous exposures

        # Getting predictions
        g[self.outcome] = np.nan
        g[self.outcome] = self._outcome_model.predict(g)
        if not predict_missing:
            g[self.outcome] = np.where(self.gf[self.outcome].isna(), np.nan, g[self.outcome])

        self.predicted_df = g

        if self._weights is None:  # unweighted marginal estimate
            g = g.dropna()
            if self.standardize == 'population':
                self.marginal_outcome = np.mean(g[self.outcome])
            elif self.standardize == 'exposed':
                self.marginal_outcome = np.mean(g.loc[self.gf[self.exposure] == 1, self.outcome])
            else:
                self.marginal_outcome = np.mean(g.loc[self.gf[self.exposure] == 0, self.outcome])
        else:  # weighted marginal estimate
            g = g.dropna()
            if self.standardize == 'population':
                self.marginal_outcome = np.average(g[self.outcome], weights=g[self._weights])
            elif self.standardize == 'exposed':
                self.marginal_outcome = np.average(g.loc[self.gf[self.exposure] == 1, self.outcome],
                                                   weights=g[self._weights])
            else:
                self.marginal_outcome = np.average(g.loc[self.gf[self.exposure] == 0, self.outcome],
                                                   weights=g[self._weights])

    def fit_stochastic(self, p, conditional=None, samples=100, predict_missing=True, seed=None):
        """Fits the g-formula for a stochastic intervention. As currently implemented, `p` percent of the population
        is randomly treated. This process is repeated `n` times and the mean is the marginal stochastic outcome.

        Parameters
        ----------
        p: float, list
            Percent of the population to randomly treat
        conditional: list
            Exclusive conditions to place on the data set for treatment percents. If specified, must match the length
            of the list of probabilities in 'p'
        samples: int, optional
            Number of resamples to calculate the marginal outcome. Default is 100
        predict_missing : bool, optional
            Whether missing outcome values should be predicted via the model fit to the data with observed outcomes.
            Default is set to True, which assumes that outcome data is missing completely at random conditional on
            the variables included in the outcome model. Note this is the less restrictive assumption regarding missing
            outcome data. If set to False, missing outcome data is assumed to be missing completely at random
            (also referred to as non-informative censoring)
        seed: int, optional
            Seed for the random process selection

        References
        ----------
        J Ahern, KE Colson, C Margerson-Zilko, A Hubbard, & S Galea. (2016). Predicting the population
        health impacts of community interventions: the case of alcohol outlets and binge drinking. American
        Journal of Public Health, 106(11), 1938-1943.

        Muñoz, ID, & van der Laan, M (2012). Population intervention causal effects based on stochastic
        interventions. Biometrics, 68(2), 541-549.
        """
        # Checking for common problems before estimation
        if self._outcome_model is None:
            raise ValueError('Before the g-formula can be calculated, the outcome model must be specified')
        if self.exposure_type != 'binary':
            raise ValueError('Only binary exposures are currently available for the stochastic treatment g-formula')
        if conditional is not None:
            if len(p) != len(conditional):
                raise ValueError("'p' and 'conditional' must be the same length")
        else:
            if type(p) != float:
                raise ValueError("Specified percent must be a float when 'conditional' is not specified")

        if seed is None:
            pass
        else:
            np.random.seed(seed)

        if conditional is not None:  # Check exclusive conditional categories
            self._check_conditional(conditional)

        marginals = []
        for s in range(samples):
            g = self.gf.reset_index(drop=True).copy()

            # Selecting out observations
            if conditional is None:  # Unconditional probabilities
                pr = g[self._weights] / np.sum(g[self._weights]) if self._weights is not None else None
                treated = np.random.choice(g.index, size=int(p*g.shape[0]), replace=False, p=pr)

            else:  # Conditional probabilities based on eval()
                treated = []
                for c, prop in zip(conditional, p):
                    gs = g.loc[eval(c)].copy()
                    pr = gs[self._weights] / np.sum(gs[self._weights]) if self._weights is not None else None
                    tr = np.random.choice(gs.index, size=int(prop*gs.shape[0]), replace=False, p=pr)
                    treated.extend(tr)

            # Applying treatment (binary only currently)
            g[self.exposure] = np.where(g.index.isin(treated), 1, 0)

            # Getting predictions
            g[self.outcome] = np.nan
            g[self.outcome] = self._outcome_model.predict(g)
            if not predict_missing:
                g[self.outcome] = np.where(self.gf[self.outcome].isna(), np.nan, g[self.outcome])
                g = g.dropna()

            if self._weights is None:  # unweighted marginal estimate
                if self.standardize == 'population':
                    marginals.append(np.mean(g[self.outcome]))
                elif self.standardize == 'exposed':
                    marginals.append(np.mean(g.loc[self.gf[self.exposure] == 1, self.outcome]))
                else:
                    marginals.append(np.mean(g.loc[self.gf[self.exposure] == 0, self.outcome]))
            else:  # weighted marginal estimate
                if self.standardize == 'population':
                    marginals.append(np.average(g[self.outcome], weights=g[self._weights]))
                elif self.standardize == 'exposed':
                    marginals.append(np.average(g.loc[self.gf[self.exposure] == 1, self.outcome],
                                                weights=g[self._weights]))
                else:
                    marginals.append(np.average(g.loc[self.gf[self.exposure] == 0, self.outcome],
                                                weights=g[self._weights]))

        self.marginal_outcome = np.mean(marginals)

    def run_diagnostics(self, decimal=3):
        """Runs diagnostics for the g-formula regression model used. Diagnostics include summary statistics and a
        Kernel Density plot for the predictive accuracy of the model. The model compares the model predicted value to
        the observed outcome value.
        """
        # Summary statistics of prediction accuracy
        outcome_accuracy(true=self.gf[self.outcome], predicted=self._predicted_y_, decimal=decimal)

        # Distribution plot of accuracy
        self.plot_kde()
        plt.title("Kernel Density of Accuracy")
        plt.tight_layout()
        plt.show()

    def plot_kde(self, bw_method='scott', fill=True, color='b'):
        """Generates a Kernel Density plot of the accuracy of the model predicted outcomes. The plot compares the
        model predicted outcome to the observed outcome. This can be used as a diagnostic for the g-formula.

        Parameters
        ----------
        bw_method : str, optional
            Method used to estimate the bandwidth. Following SciPy, either 'scott' or 'silverman' are valid options
        fill : bool, optional
            Whether to color the area under the density curves. Default is true
        color : str, optional
            Color of the line/area. Default is blue

        Returns
        -------
        matplotlib axes
        """
        if self._predicted_y_ is None:
            raise ValueError("The outcome_model function must be ran before any diagnostics")

        v = self._predicted_y_ - self.gf[self.outcome]
        return plot_kde_accuracy(values=v.dropna(),
                                 bw_method=bw_method, fill=fill, color=color)

    def _check_conditional(self, conditional):
        """Check that conditionals are exclusive for the stochastic fit process
        """
        g = self.gf.copy()
        a = np.array([0]*g.shape[0])
        for c in conditional:
            a = np.add(a, np.where(eval(c), 1, 0))

        if np.sum(np.where(a > 1, 1, 0)):
            warnings.warn("It looks like your conditional categories are NOT exclusive. For appropriate estimation, "
                          "the conditions that designate each category should be exclusive", UserWarning)


class SurvivalGFormula:
    r"""G-formula for time-to-event data where the exposure is fixed at baseline. Only supports binary exposures and
    outcomes. Outcomes are predicted using a logistic regression model. Input data set should be in a long format,
    where each row corresponds to an individual observed for one unit of time

    Key options for treatments:

        * `'all'`     -all individuals are given treatment
        * `'none'`    -no individuals are given treatment
        * custom treatments -create a custom treatment. When specifying this, the dataframe must be referred to as 'g'.
          The following is an example that selects those whose age is 30 or younger and are females:
          ``treatment="((g['age0']<=30) & (g['male']==0))``

    Parameters
    ----------
    df : DataFrame
        Pandas dataframe containing the variables of interest
    idvar : str
        Column name for the ID label
    exposure : str, list
        Column name for exposure variable label or a list of disjoint indicator exposures
    outcome : str
        Column name for outcome variable
    time : str
        Column name for time variable
    weights : str, optional
        Column name for weights. Default is None, which assumes every observations has the same weight (i.e. 1)

    Note
    ----
    Custom treatments use a "magic-g" parameter. Internally, the g-formula implementation names the data set as `g`.
    Therefore, when using custom treatment specifications, the data set must be referred to as `g` when following
    the pandas selection syntax

    Examples
    --------
    Setting up data in long format

    >>> from zepid import load_sample_data
    >>> from zepid.causal.gformula import SurvivalGFormula
    >>> import matplotlib.pyplot as plt
    >>> df = load_sample_data(False).drop(columns=['cd4_wk45'])

    >>> df['t'] = np.round(df['t']).astype(int)
    >>> df = pd.DataFrame(np.repeat(df.values, df['t'], axis=0), columns=df.columns)
    >>> df['t'] = df.groupby('id')['t'].cumcount() + 1
    >>> df.loc[((df['dead'] == 1) & (df['id'] != df['id'].shift(-1))), 'd'] = 1
    >>> df['d'] = df['d'].fillna(0)
    >>> df['t_sq'] = df['t']**2
    >>> df['t_cu'] = df['t']**3

    Estimating the time-to-event mean effect under treat-all plan

    >>> sgf = SurvivalGFormula(df.drop(columns=['dead']), idvar='id', exposure='art', outcome='d', time='t')
    >>> sgf.outcome_model(model='art + male + age0 + cd40 + dvl0 + t + t_sq + t_cu')
    >>> sgf.fit(treatment='all')
    >>> print(sgf.marginal_outcome)

    Plotting cumulative incidence function

    >>> sgf.plot(color='r')
    >>> plt.show()

    Estimating the time-to-event mean effect under treat-none plan

    >>> sgf = SurvivalGFormula(df.drop(columns=['dead']), idvar='id', exposure='art', outcome='d', time='t')
    >>> sgf.outcome_model(model='art + male + age0 + cd40 + dvl0 + t + t_sq + t_cu')
    >>> sgf.fit(treatment='none')

    Estimating the time-to-event mean effect under custom treatment plan

    >>> sgf = SurvivalGFormula(df.drop(columns=['dead']), idvar='id', exposure='art', outcome='d', time='t')
    >>> sgf.outcome_model(model='art + male + age0 + cd40 + dvl0 + t + t_sq + t_cu')
    >>> sgf.fit(treatment="((g['age0']>=25) & (g['male']==0))")

    Notes
    -----
    The following process is used to estimate the cumulative incidence function.
    (1) A pooled logistic regression model is fit to the data. The model should predict the outcome conditional on
    treatment, baseline confounders, and time. Time should be modeled using flexible functional forms (e.g. splines)
    (2) Survival probabilities are estimated by predicting values at each time from the pooled logistic model and taking
    the cumulative product. The survival probabilities are predicted under the treatment plan of interest
    (3) Average the cumulative incidence function for each time period from all the subjects.

    References
    ----------
    Hernán MA. (2010). The hazards of hazard ratios. Epidemiology, 21(1), 13–15.
    doi:10.1097/EDE.0b013e3181c1ea43
    """
    def __init__(self, df, idvar, exposure, outcome, time, weights=None):
        self.exposure = exposure
        self.outcome = outcome
        self.t = time
        self.id = idvar
        self._missing_indicator = '__missing_indicator__'
        self.gf, self._miss_flag, self._continuous_outcome_ = check_input_data(data=df,
                                                                               exposure=exposure,
                                                                               outcome=outcome,
                                                                               estimator="SurvivalGFormula",
                                                                               drop_censoring=True,
                                                                               drop_missing=True,
                                                                               binary_exposure_only=True)
        self.gf = self.gf.copy().sort_values(by=[idvar, time]).reset_index(drop=True)
        if self._continuous_outcome_:
            raise ValueError("SurvivalGFormula does not support continuous outcomes")

        self._weights = weights
        self._outcome_model = None
        self.marginal_outcome = None
        self.predicted_df = None

    def outcome_model(self, model, print_results=True):
        """Build the pooled logistic model. This must be specified before the fit function. It is encouraged to model
        time as flexible as possible. For example, use spline terms

        Parameters
        ----------
        model : str
            Variables to include in the model for predicting the outcome. Must be contained within the input
            pandas dataframe when initialized. Model form should contain the exposure, i.e. 'art + age + male'
        print_results : bool, optional
            Whether to print the logistic regression results to the terminal. Default is True
        """
        if self.exposure not in model:
            warnings.warn("It looks like '" + self.exposure + "' is not included in the outcome model.")

        # Modeling the outcome
        linkdist = sm.families.family.Binomial()
        if self._weights is None:
            m = smf.glm(self.outcome + ' ~ ' + model, self.gf, family=linkdist)
        else:
            m = smf.glm(self.outcome + ' ~ ' + model, self.gf, family=linkdist,
                        freq_weights=self.gf[self._weights])

        self._outcome_model = m.fit()

        # Printing results of the model
        if print_results:
            print('==============================================================================')
            print('Outcome Model')
            print(self._outcome_model.summary())
            print('==============================================================================')

    def fit(self, treatment):
        """Fit the parametric g-formula for time-to-event data. To obtain the confidence intervals, use a bootstrap
        procedure

        Parameters
        ----------
        treatment : str, list
            There are three options available for treatment plans. All, none, or a custom pattern
              * `'all'`     -all individuals are given treatment
              * `'none'`    -no individuals are given treatment
              * `'natural'` -all individuals retain their observed exposure
              * custom  -create a custom treatment. When specifying this, the dataframe must be referred to as 'g' The
                following is an example that selects those whose age is 25 or older and are females;
                ``treatment="((g['age0']>=25) & (g['male']==0))``

        Returns
        -------
        marginal_outcome
            Parameter of `marginal_outcome` is filled, which is the mean predicted outcome under the treatment strategy
        """
        if self._outcome_model is None:
            raise ValueError('Before the g-formula can be calculated, the outcome model must be specified')

        # Setting outcome as blank
        g = self.gf.copy()

        # Setting treatment (either multivariate or binary)
        if treatment == 'all':
            g[self.exposure] = 1
        elif treatment == 'none':
            g[self.exposure] = 0
        elif treatment == 'natural':
            pass
        else:  # custom exposure pattern
            g[self.exposure] = np.where(eval(treatment), 1, 0)

        g[self.outcome] = np.nan
        g[self.outcome] = 1 - self._outcome_model.predict(g)

        # Cumulative product
        g[self.outcome] = 1 - g.groupby(self.id)[self.outcome].cumprod()

        if self._weights is None:  # unweighted marginal estimate
            marginal = g.groupby(self.t)[self.outcome].mean()

        else:  # weighted marginal estimate
            marginal = self._weighted_average(data=g, y_col=self.outcome, weight_col=self._weights, by_col=self.t)

        self.marginal_outcome = marginal.rename(index='timeline')
        self.predicted_df = g

    def plot(self, **plot_kwargs):
        """Plots the estimated cumulative incidence function

        Parameters
        ----------
        **plot_kwargs : optional
            Optional keyword arguments for matplotlib. kwargs will pass matplotlib.pyploy.step kwargs are accepted. See
            matplotlib 'step()' function documentation for further details

        Returns
        -------
        matplotlib axes
        """
        if self.marginal_outcome is None:
            raise ValueError('Before plotting, the marginal outcomes must be estimated with the fit() function')

        ax = plt.gca()
        ax.step(self.marginal_outcome.index, self.marginal_outcome, where='post', **plot_kwargs)
        ax.set_xlabel(self.t)
        ax.set_ylabel('Risk of ' + self.outcome)
        return ax

    @staticmethod
    def _weighted_average(data, y_col, weight_col, by_col):
        """Background function to calculate the weighted mean by time
        """
        data['_w_y_'] = data[y_col] * data[weight_col]
        data['_w_t_'] = data[weight_col] * pd.notnull(data[y_col])
        g = data.groupby(by_col)
        result = g['_w_y_'].sum() / g['_w_t_'].sum()
        return result
